from eps_spine_shared.common.prescription import fields
from eps_spine_shared.common.prescription.statuses import LineItemStatus


class PrescriptionLineItem(object):
    """
    Wrapper class to simplify interacting with line item sections of a prescription record.
    """

    def __init__(self, line_item_dict):
        """
        Constructor.

        :type line_item_dict: dict
        """
        self._line_item_dict = line_item_dict

    @property
    def id(self):
        """
        The line item's ID.

        :rtype: str
        """
        return self._line_item_dict[fields.FIELD_ID]

    @property
    def status(self):
        """
        The status of this line item.

        :rtype: str
        """
        return self._line_item_dict[fields.FIELD_STATUS]

    @property
    def previous_status(self):
        """
        The previous status of this line item.

        :rtype: str
        """
        return self._line_item_dict[fields.FIELD_PREVIOUS_STATUS]

    @property
    def order(self):
        """
        The order of this line item.

        :rtype: int
        """
        return self._line_item_dict[fields.FIELD_ORDER]

    @property
    def max_repeats(self):
        """
        The maximum number of repeats for this line item.

        :rtype: int
        """
        return int(self._line_item_dict[fields.FIELD_MAX_REPEATS])

    def is_active(self):
        """
        Test whether this line item is active.

        :rtype: bool
        """
        return self.status in LineItemStatus.ACTIVE_STATES

    def update_status(self, new_status):
        """
        Set the line item status, and remember the previous status.

        :type new_status: str
        """
        self._line_item_dict[fields.FIELD_PREVIOUS_STATUS] = self._line_item_dict[
            fields.FIELD_STATUS
        ]
        self._line_item_dict[fields.FIELD_STATUS] = new_status

    def expire(self, parent_prescription):
        """
        Expire this line item.

        :type parent_prescription: PrescriptionRecord
        """
        current_status = self.status
        if current_status not in LineItemStatus.EXPIRY_IMMUTABLE_STATES:
            new_status = LineItemStatus.EXPIRY_LOOKUP[current_status]
            self.update_status(new_status)
            parent_prescription.log_object.write_log(
                "EPS0072b",
                None,
                {
                    "internalID": parent_prescription.internal_id,
                    "lineItemChanged": self.id,
                    "previousStatus": current_status,
                    "newStatus": new_status,
                },
            )



from enum import Enum

from botocore.exceptions import NoCredentialsError

# Try to import spine error classes. If successful, we are on spine and should use wrapper classes.
on_spine = False
try:
    # from spinecore.prescriptions.common.errors.errorbaseprescriptionsearch \
    #     import ErrorBasePrescSearch # pyright: ignore[reportMissingImports]
    from spinecore.common.aws.awscommon import (  # pyright: ignore[reportMissingImports]
        NoCredentialsErrorWithRetry,
    )
    from spinecore.common.errors import (  # pyright: ignore[reportMissingImports]
        SpineBusinessError,
        SpineSystemError,
    )

    # from spinecore.prescriptions.common.errors.errorbase1634 \
    #     import ErrorBase1634 # pyright: ignore[reportMissingImports]
    from spinecore.prescriptions.common.errors.errorbase1719 import (  # pyright: ignore[reportMissingImports]
        ErrorBase1719,
    )
    from spinecore.prescriptions.common.errors.errorbase1722 import (  # pyright: ignore[reportMissingImports]
        ErrorBase1722,
    )

    on_spine = True
except ImportError:
    pass


if on_spine:

    class EpsNoCredentialsErrorWithRetry:
        """
        Wrapper for NoCredentialsErrorWithRetry
        """

        def __init__(self, *args):
            raise NoCredentialsErrorWithRetry(*args)

    class EpsSystemError:
        """
        Wrapper for SpineSystemError
        """

        def __init__(self, *args):
            raise SpineSystemError(*args)

    class EpsBusinessError:
        """
        Wrapper for SpineBusinessError
        """

        def __init__(self, *args):
            raise SpineBusinessError(*args)

    class EpsErrorBase:
        """
        Wrapper for ErrorBases
        """

        MISSING_ISSUE = ErrorBase1722.PRESCRIPTION_NOT_FOUND
        ITEM_NOT_FOUND = ErrorBase1722.ITEM_NOT_FOUND
        INVALID_LINE_STATE_TRANSITION = ErrorBase1722.INVALID_LINE_STATE_TRANSITION
        MAX_REPEAT_MISMATCH = ErrorBase1722.MAX_REPEAT_MISMATCH
        NOT_CANCELLED_EXPIRED = ErrorBase1719.NOT_CANCELLED_EXPIRED
        NOT_CANCELLED_CANCELLED = ErrorBase1719.NOT_CANCELLED_CANCELLED
        NOT_CANCELLED_NOT_DISPENSED = ErrorBase1719.NOT_CANCELLED_NOT_DISPENSED
        NOT_CANCELLED_DISPENSED = ErrorBase1719.NOT_CANCELLED_DISPENSED
        NOT_CANCELLED_WITH_DISPENSER = ErrorBase1719.NOT_CANCELLED_WITH_DISPENSER
        NOT_CANCELLED_WITH_DISPENSER_ACTIVE = ErrorBase1719.NOT_CANCELLED_WITH_DISPENSER_ACTIVE
        PRESCRIPTION_NOT_FOUND = ErrorBase1719.PRESCRIPTION_NOT_FOUND

else:

    class EpsNoCredentialsErrorWithRetry(NoCredentialsError):
        """
        Extends NoCredentialsError to provide information about retry attempts.
        """

        fmt = "Unable to locate credentials after {attempts} attempts"

    class EpsSystemError(Exception):
        """
        Exception to be raised if an unexpected system error occurs.
        """

        MESSAGE_FAILURE = "messageFailure"
        DEVELOPMENT_FAILURE = "developmentFailure"
        SYSTEM_FAILURE = "systemFailure"
        IMMEDIATE_REQUEUE = "immediateRequeue"
        RETRY_EXPIRED = "retryExpired"
        PUBLISHER_HANDLES_REQUEUE = "publisherHandlesRequeue"
        UNRELIABLE_MESSAGE = "unreliableMessage"

        def __init__(self, error_topic, *args):  # noqa: B042
            """
            error_topic is the topic to be used when writing the WDO to the error exchange
            """
            super(EpsSystemError, self).__init__(*args)
            self.error_topic = error_topic

    class EpsBusinessError(Exception):
        """
        Exception to be raised by a message worker if an expected error condition is hit,
        one that is expected to cause a HL7 error response with a set errorCode.
        """

        def __init__(self, error_code, supp_info=None, message_id=None):  # noqa: B042
            super(EpsBusinessError, self).__init__()
            self.error_code = error_code
            self.supplementary_information = supp_info
            self.message_id = message_id

        def __str__(self):
            if self.supplementary_information:
                return "{} {}".format(self.error_code, self.supplementary_information)
            return str(self.error_code)

    class EpsErrorBase(Enum):
        """
        To be used in Spine application code to remap to ErrorBases.
        """

        INVALID_LINE_STATE_TRANSITION = 1
        ITEM_NOT_FOUND = 2
        MAX_REPEAT_MISMATCH = 3
        NOT_CANCELLED_EXPIRED = 4
        NOT_CANCELLED_CANCELLED = 5
        NOT_CANCELLED_NOT_DISPENSED = 6
        NOT_CANCELLED_DISPENSED = 7
        NOT_CANCELLED_WITH_DISPENSER = 8
        NOT_CANCELLED_WITH_DISPENSER_ACTIVE = 9
        PRESCRIPTION_NOT_FOUND = 10
        MISSING_ISSUE = 11



from eps_spine_shared.common.prescription import fields
from eps_spine_shared.common.prescription.record import PrescriptionRecord
from eps_spine_shared.spinecore.baseutilities import quoted


class SinglePrescribeRecord(PrescriptionRecord):
    """
    Class defined to handle single instance (acute) prescriptions
    """

    def __init__(self, log_object, internal_id):
        """
        Allow the record_type attribute to be set
        """
        super(SinglePrescribeRecord, self).__init__(log_object, internal_id)
        self.record_type = "Acute"

    def add_line_item_repeat_data(self, release_data, line_item_ref, line_item):
        """
        Add line item information (This is not required for Acute prescriptions, but
        will invalidate the signature if provided in the prescription and not returned
        in the release.
        It the lineItem.max_repeats is false (not provided inbound), then do not include
        it in the response, otherwise, both MaxRepeats and CurrentInstnace will be 1 for Acute.

        :type release_data: dict
        :type line_item_ref: str
        :type line_item: PrescriptionLineItem
        """
        # Handle the missing inbound max_repeats
        if not line_item.max_repeats:
            return

        # Acute, so both values may only be '1'
        release_data[line_item_ref + "MaxRepeats"] = quoted(str(1))
        release_data[line_item_ref + "CurrentInstance"] = quoted(str(1))

    def return_details_for_dispense(self):
        """
        For dispense messages the following details are required:
        - Issue number
        - Issue status
        - NHS Number
        - Dispensing Organisation
        - None (indicating not a repeat prescription so no max_repeats)
        """
        current_issue = self.current_issue
        details = [
            str(current_issue.number),
            current_issue.status,
            self._nhs_number,
            current_issue.dispensing_organization,
            None,
        ]
        return details

    def return_details_for_claim(self, instance_number_str):
        """
        For dispense messages the following details are required:
        - Issue status
        - NHS Number
        - Dispensing Organisation
        - None (indicating not a repeat prescription so no max_repeats)
        """
        issue_number = int(instance_number_str)
        issue = self.get_issue(issue_number)
        details = [
            issue.claim,
            issue.status,
            self._nhs_number,
            issue.dispensing_organization,
            None,
        ]
        return details

    def return_last_dispense_date(self, instance_number):
        """
        Return the last_dispense_date for the requested instance
        """
        instance = self._get_prescription_instance_data(instance_number)
        last_dispense_date = instance[fields.FIELD_DISPENSE][fields.FIELD_LAST_DISPENSE_DATE]
        return last_dispense_date

    def return_last_disp_msg_ref(self, instance_number_str):
        """
        returns the last dispense Msg Ref for the issue
        """
        issue_number = int(instance_number_str)
        issue = self.get_issue(issue_number)
        return issue.last_dispense_notification_msg_ref



import re
from datetime import datetime, timedelta
from typing import Tuple

from boto3.dynamodb.conditions import Attr
from boto3.dynamodb.conditions import Key as BotoKey

from eps_spine_shared.common import indexes
from eps_spine_shared.common.dynamodb_client import EpsDynamoDbClient
from eps_spine_shared.common.dynamodb_common import (
    GSI,
    NEXT_ACTIVITY_DATE_PARTITIONS,
    Attribute,
    Key,
    ProjectedAttribute,
    SortKey,
)
from eps_spine_shared.common.prescription.record import PrescriptionStatus
from eps_spine_shared.logger import EpsLogger
from eps_spine_shared.nhsfundamentals.timeutilities import TimeFormats


class EpsDynamoDbIndex:
    """
    The prescriptions message store specific DynamoDB client.
    """

    def __init__(self, log_object, client: EpsDynamoDbClient):
        """
        Instantiate the DynamoDB client.
        """
        self.log_object = EpsLogger(log_object)
        self.client = client

    def nhs_number_date(self, range_start, range_end, term_regex):
        """
        Query the nhsNumberDate index.
        """
        # POC - Use context in these methods, rather than range_start and range_end.
        nhs_number, start_date = range_start.split(indexes.SEPERATOR)
        end_date = range_end.split(indexes.SEPERATOR)[-1]

        return self.query_nhs_number_date(
            indexes.INDEX_NHSNUMBER_DATE, nhs_number, start_date, end_date, term_regex=term_regex
        )

    def nhs_number_presc_disp_date(self, range_start, range_end, term_regex):
        """
        Query the nhsNumberDate index, filtering on prescriber and dispenser.
        """
        nhs_number, prescriber_org, dispenser_org, start_date = range_start.split(indexes.SEPERATOR)
        end_date = range_end.split(indexes.SEPERATOR)[-1]
        filter_expression = Attr(Attribute.PRESCRIBER_ORG.name).eq(prescriber_org) & Attr(
            Attribute.DISPENSER_ORG.name
        ).contains(dispenser_org)

        return self.query_nhs_number_date(
            indexes.INDEX_NHSNUMBER_PRDSDATE,
            nhs_number,
            start_date,
            end_date,
            filter_expression,
            term_regex,
        )

    def nhs_number_presc_date(self, range_start, range_end, term_regex):
        """
        Query the nhsNumberDate index, filtering on prescriber.
        """
        nhs_number, prescriber_org, start_date = range_start.split(indexes.SEPERATOR)
        end_date = range_end.split(indexes.SEPERATOR)[-1]
        filter_expression = Attr(Attribute.PRESCRIBER_ORG.name).eq(prescriber_org)

        return self.query_nhs_number_date(
            indexes.INDEX_NHSNUMBER_PRDATE,
            nhs_number,
            start_date,
            end_date,
            filter_expression,
            term_regex,
        )

    def nhs_number_disp_date(self, range_start, range_end, term_regex):
        """
        Query the nhsNumberDate index, filtering on dispenser.
        """
        nhs_number, dispenser_org, start_date = range_start.split(indexes.SEPERATOR)
        end_date = range_end.split(indexes.SEPERATOR)[-1]
        filter_expression = Attr(Attribute.DISPENSER_ORG.name).contains(dispenser_org)

        return self.query_nhs_number_date(
            indexes.INDEX_NHSNUMBER_DSDATE,
            nhs_number,
            start_date,
            end_date,
            filter_expression,
            term_regex,
        )

    def presc_disp_date(self, range_start, range_end, term_regex):
        """
        Query the prescriberDate index, filtering on dispenser.
        """
        prescriber_org, dispenser_org, start_date = range_start.split(indexes.SEPERATOR)
        end_date = range_end.split(indexes.SEPERATOR)[-1]
        filter_expression = Attr(Attribute.DISPENSER_ORG.name).contains(dispenser_org)

        return self.query_prescriber_date(
            indexes.INDEX_PRESCRIBER_DSDATE,
            prescriber_org,
            start_date,
            end_date,
            filter_expression,
            term_regex,
        )

    def presc_date(self, range_start, range_end, term_regex):
        """
        Query the prescriberDate index.
        """
        prescriber_org, start_date = range_start.split(indexes.SEPERATOR)
        end_date = range_end.split(indexes.SEPERATOR)[-1]

        return self.query_prescriber_date(
            indexes.INDEX_PRESCRIBER_DATE,
            prescriber_org,
            start_date,
            end_date,
            term_regex=term_regex,
        )

    def disp_date(self, range_start, range_end, term_regex):
        """
        Query the dispenserDate index.
        """
        dispenser_org, start_date = range_start.split(indexes.SEPERATOR)
        end_date = range_end.split(indexes.SEPERATOR)[-1]

        return self.query_dispenser_date(
            indexes.INDEX_DISPENSER_DATE, dispenser_org, start_date, end_date, term_regex=term_regex
        )

    def nom_pharm_status(self, range_start, _, term_regex):
        """
        Query the nomPharmStatus index for terms.
        """
        ods_code, status = range_start.split("_")

        return self.query_nom_pharm_status_terms(
            indexes.INDEX_NOMPHARM, ods_code, status, term_regex=term_regex
        )

    def build_terms(self, items, index_name, term_regex):
        """
        Build terms from items returned by the index query.
        """
        terms = []
        for item in items:
            index_terms = item.get(ProjectedAttribute.INDEXES.name, {}).get(index_name.lower())
            if not index_terms:
                continue
            [
                terms.append((index_term, item[Key.PK.name]))
                for index_term in index_terms
                if ((not term_regex) or re.search(term_regex, index_term))
            ]
        return terms

    def pad_or_trim_date(self, date):
        """
        Ensure the date length is fourteen characters, if present.
        """
        if not date:
            return None

        if len(date) >= 14:
            return date[:14]

        while len(date) < 14:
            date = date + "0"
        return date

    def query_nhs_number_date(
        self,
        index,
        nhs_number,
        start_date=None,
        end_date=None,
        filter_expression=None,
        term_regex=None,
    ):
        """
        Return the epsRecord terms which match the supplied range and regex for the nhsNumberDate index.
        """
        start_date, end_date = [self.pad_or_trim_date(date) for date in [start_date, end_date]]

        pk_expression = BotoKey(Attribute.NHS_NUMBER.name).eq(nhs_number)
        sk_expression = None
        if start_date and end_date:
            [valid, sk_expression] = self._get_valid_range_condition(
                Attribute.CREATION_DATETIME.name, start_date, end_date
            )

            if not valid:
                return []
        elif start_date:
            sk_expression = BotoKey(Attribute.CREATION_DATETIME.name).gte(start_date)
        elif end_date:
            sk_expression = BotoKey(Attribute.CREATION_DATETIME.name).lte(end_date)

        key_condition_expression = (
            pk_expression if not sk_expression else pk_expression & sk_expression
        )
        items = self.client.query_index(
            GSI.NHS_NUMBER_DATE.name, key_condition_expression, filter_expression
        )

        return self.build_terms(items, index, term_regex)

    def query_prescriber_date(
        self, index, prescriber_org, start_date, end_date, filter_expression=None, term_regex=None
    ):
        """
        Return the epsRecord terms which match the supplied range and regex for the prescriberDate index.
        """
        start_date, end_date = [self.pad_or_trim_date(date) for date in [start_date, end_date]]

        pk_expression = BotoKey(Attribute.PRESCRIBER_ORG.name).eq(prescriber_org)
        [valid, sk_expression] = self._get_valid_range_condition(
            Attribute.CREATION_DATETIME.name, start_date, end_date
        )

        if not valid:
            return []

        items = self.client.query_index(
            GSI.PRESCRIBER_DATE.name, pk_expression & sk_expression, filter_expression
        )

        return self.build_terms(items, index, term_regex)

    def query_dispenser_date(
        self, index, dispenser_org, start_date, end_date, filter_expression=None, term_regex=None
    ):
        """
        Return the epsRecord terms which match the supplied range and regex for the dispenserDate index.
        """
        start_date, end_date = [self.pad_or_trim_date(date) for date in [start_date, end_date]]

        pk_expression = BotoKey(Attribute.DISPENSER_ORG.name).eq(dispenser_org)
        [valid, sk_expression] = self._get_valid_range_condition(
            Attribute.CREATION_DATETIME.name, start_date, end_date
        )

        if not valid:
            return []

        items = self.client.query_index(
            GSI.DISPENSER_DATE.name, pk_expression & sk_expression, filter_expression
        )

        return self.build_terms(items, index, term_regex)

    def query_nom_pharm_status(self, ods_code, all_statuses=False, limit=None):
        """
        Return the nomPharmStatus prescription keys which match the supplied ODS code.
        Query using the nominatedPharmacyStatus index. If all_statuses is False, only return prescriptions
        with status TO_BE_DISPENSED (0001).
        """
        key_condition_expression = BotoKey(Attribute.NOMINATED_PHARMACY.name).eq(ods_code)

        is_ready_condition = (
            BotoKey(Attribute.IS_READY.name).eq(int(True))
            if not all_statuses
            else BotoKey(Attribute.IS_READY.name).between(0, 1)
        )
        key_condition_expression = key_condition_expression & is_ready_condition

        items = self.client.query_index_with_limit(
            GSI.NOMINATED_PHARMACY_STATUS.name, key_condition_expression, None, limit
        )

        return [item[Key.PK.name] for item in items]

    def query_nom_pharm_status_terms(self, index, ods_code, status, term_regex=None):
        """
        Return the nomPharmStatus terms which match the supplied ODS code and status.
        Query using the nominatedPharmacyStatus index, with is_ready derived from the status.
        """
        is_ready = status == PrescriptionStatus.TO_BE_DISPENSED

        key_condition_expression = BotoKey(Attribute.NOMINATED_PHARMACY.name).eq(
            ods_code
        ) & BotoKey(Attribute.IS_READY.name).eq(int(is_ready))

        filter_expression = Attr(ProjectedAttribute.STATUS.name).contains(status)

        items = self.client.query_index(
            GSI.NOMINATED_PHARMACY_STATUS.name, key_condition_expression, filter_expression
        )

        return self.build_terms(items, index, term_regex)

    def query_claim_id(self, claim_id):
        """
        Search for an existing batch claim containing the given claim_id.
        """
        key_condition_expression = BotoKey(Key.SK.name).eq(SortKey.CLAIM.value)
        filter_expression = Attr(ProjectedAttribute.CLAIM_IDS.name).contains(claim_id)

        items = self.client.query_index(
            GSI.CLAIM_ID.name, key_condition_expression, filter_expression
        )

        return [item[Key.PK.name] for item in items]

    def query_next_activity_date(self, range_start, range_end, shard=None):
        """
        Yields the epsRecord keys which match the supplied nextActivity and date range for the nextActivity index.

        nextActivity is suffix-sharded with NEXT_ACTIVITY_DATE_PARTITIONS to avoid hot partitions on ddb.
        This means NEXT_ACTIVITY_DATE_PARITIONS + 1 queries are performed, one for each partition
        and one for the non-partitioned nextActivityDate index.
        """
        next_activity, start_date = range_start.split("_")
        end_date = range_end.split("_")[-1]

        [valid, sk_expression] = self._get_valid_range_condition(
            Attribute.NEXT_ACTIVITY_DATE.name, start_date, end_date
        )

        if not valid:
            return []

        if shard or shard == "":
            yield from self._query_next_activity_date_shard(next_activity, sk_expression, shard)
            return

        shards = [None] + list(range(1, NEXT_ACTIVITY_DATE_PARTITIONS + 1))

        for shard in shards:
            yield from self._query_next_activity_date_shard(next_activity, sk_expression, shard)

    def _query_next_activity_date_shard(self, next_activity, sk_expression, shard):
        """
        Return a generator for the epsRecord keys which match the supplied nextActivity and date range
        for a given pk shard.
        """
        expected_next_activity = (
            next_activity if shard is None or shard == "" else f"{next_activity}.{shard}"
        )
        pk_expression = BotoKey(Attribute.NEXT_ACTIVITY.name).eq(expected_next_activity)

        return self.client.query_index_yield(
            GSI.NEXT_ACTIVITY_DATE.name, pk_expression & sk_expression
        )

    def _get_date_range_for_query(self, start_datetime_str, end_datetime_str):
        """
        Get days included in the given range. For use in claimNotificationStoreTime index query.
        """
        start_datetime = datetime.strptime(
            start_datetime_str, TimeFormats.STANDARD_DATE_TIME_FORMAT
        )
        end_datetime = datetime.strptime(end_datetime_str, TimeFormats.STANDARD_DATE_TIME_FORMAT)

        return [
            (start_datetime + timedelta(days=d)).strftime(TimeFormats.STANDARD_DATE_FORMAT)
            for d in range((end_datetime.date() - start_datetime.date()).days + 1)
        ]

    def query_claim_notification_store_time(
        self, internal_id, start_datetime_str, end_datetime_str
    ):
        """
        Search for claim notification documents whose store times fall within the specified window.
        """
        [valid, sk_expression] = self._get_valid_range_condition(
            Attribute.STORE_TIME.name, start_datetime_str, end_datetime_str
        )

        if not valid:
            return []

        dates = self._get_date_range_for_query(start_datetime_str, end_datetime_str)
        generators = []

        for date in dates:
            pk_expression = BotoKey(Attribute.CLAIM_NOTIFICATION_STORE_DATE.name).eq(date)
            self.log_object.write_log(
                "DDB0013",
                None,
                {
                    "date": date,
                    "startTime": start_datetime_str,
                    "endTime": end_datetime_str,
                    "internalID": internal_id,
                },
            )
            generators.append(
                self.client.query_index_yield(
                    GSI.CLAIM_NOTIFICATION_STORE_TIME.name, pk_expression & sk_expression, None
                )
            )

        for generator in generators:
            yield from generator

    def _get_valid_range_condition(self, key, start, end) -> Tuple[bool, object]:
        """
        Returns a range condition if the start < end
        """
        if end == start:
            return True, BotoKey(key).eq(start)
        if end < start:
            return False, None
        else:
            return True, BotoKey(key).between(start, end)

    def query_batch_claim_id_sequence_number(self, sequence_number, nwssp=False):
        """
        Query the claimIdSequenceNumber index for batch claim IDs based on sequence number.
        """
        index_name = (
            GSI.CLAIM_ID_SEQUENCE_NUMBER_NWSSP.name if nwssp else GSI.CLAIM_ID_SEQUENCE_NUMBER.name
        )
        key_name = Attribute.SEQUENCE_NUMBER_NWSSP.name if nwssp else Attribute.SEQUENCE_NUMBER.name

        key_condition_expression = BotoKey(key_name).eq(sequence_number)

        items = self.client.query_index(index_name, key_condition_expression, None)

        return [
            item[Key.PK.name]
            for item in items
            if item[Key.PK.name] not in ["claimSequenceNumber", "claimSequenceNumberNwssp"]
        ]



import datetime

from eps_spine_shared.common.prescription import fields
from eps_spine_shared.common.prescription.statuses import PrescriptionStatus
from eps_spine_shared.logger import EpsLogger
from eps_spine_shared.nhsfundamentals.timeutilities import TimeFormats


class NextActivityGenerator(object):
    """
    Used to create the next activity for a prescription instance
    """

    INPUT_LIST_1 = [
        fields.FIELD_EXPIRY_PERIOD,
        fields.FIELD_PRESCRIPTION_DATE,
        fields.FIELD_NOMINATED_DOWNLOAD_DATE,
        fields.FIELD_DISPENSE_WINDOW_HIGH_DATE,
    ]
    INPUT_LIST_2 = [
        fields.FIELD_EXPIRY_PERIOD,
        fields.FIELD_PRESCRIPTION_DATE,
        fields.FIELD_DISPENSE_WINDOW_HIGH_DATE,
        fields.FIELD_LAST_DISPENSE_DATE,
    ]
    INPUT_LIST_3 = [
        fields.FIELD_EXPIRY_PERIOD,
        fields.FIELD_PRESCRIPTION_DATE,
        fields.FIELD_COMPLETION_DATE,
    ]
    INPUT_LIST_4 = [
        fields.FIELD_EXPIRY_PERIOD,
        fields.FIELD_PRESCRIPTION_DATE,
        fields.FIELD_COMPLETION_DATE,
        fields.FIELD_DISPENSE_WINDOW_HIGH_DATE,
        fields.FIELD_LAST_DISPENSE_DATE,
        fields.FIELD_CLAIM_SENT_DATE,
    ]
    INPUT_LIST_5 = [
        fields.FIELD_PRESCRIBING_SITE_TEST_STATUS,
        fields.FIELD_PRESCRIPTION_DATE,
        fields.FIELD_CLAIM_SENT_DATE,
    ]
    INPUT_LIST_6 = [
        fields.FIELD_EXPIRY_PERIOD,
        fields.FIELD_PRESCRIPTION_DATE,
        fields.FIELD_NOMINATED_DOWNLOAD_DATE,
        fields.FIELD_DISPENSE_WINDOW_LOW_DATE,
    ]
    INPUT_LIST_7 = [
        fields.FIELD_EXPIRY_PERIOD,
        fields.FIELD_PRESCRIPTION_DATE,
    ]

    INPUT_BY_STATUS = {}
    INPUT_BY_STATUS[PrescriptionStatus.TO_BE_DISPENSED] = INPUT_LIST_1
    INPUT_BY_STATUS[PrescriptionStatus.WITH_DISPENSER] = INPUT_LIST_1
    INPUT_BY_STATUS[PrescriptionStatus.WITH_DISPENSER_ACTIVE] = INPUT_LIST_2
    INPUT_BY_STATUS[PrescriptionStatus.EXPIRED] = INPUT_LIST_3
    INPUT_BY_STATUS[PrescriptionStatus.CANCELLED] = INPUT_LIST_3
    INPUT_BY_STATUS[PrescriptionStatus.DISPENSED] = INPUT_LIST_4
    INPUT_BY_STATUS[PrescriptionStatus.NOT_DISPENSED] = INPUT_LIST_3
    INPUT_BY_STATUS[PrescriptionStatus.CLAIMED] = INPUT_LIST_5
    INPUT_BY_STATUS[PrescriptionStatus.NO_CLAIMED] = INPUT_LIST_5
    INPUT_BY_STATUS[PrescriptionStatus.AWAITING_RELEASE_READY] = INPUT_LIST_6
    INPUT_BY_STATUS[PrescriptionStatus.REPEAT_DISPENSE_FUTURE_INSTANCE] = INPUT_LIST_7
    INPUT_BY_STATUS[PrescriptionStatus.FUTURE_DATED_PRESCRIPTION] = INPUT_LIST_6
    INPUT_BY_STATUS[PrescriptionStatus.PENDING_CANCELLATION] = [fields.FIELD_PRESCRIPTION_DATE]

    FIELD_REPEAT_DISPENSE_EXPIRY_PERIOD = "repeatDispenseExpiryPeriod"
    FIELD_PRESCRIPTION_EXPIRY_PERIOD = "prescriptionExpiryPeriod"
    FIELD_WITH_DISPENSER_ACTIVE_EXPIRY_PERIOD = "withDispenserActiveExpiryPeriod"
    FIELD_EXPIRED_DELETE_PERIOD = "expiredDeletePeriod"
    FIELD_CANCELLED_DELETE_PERIOD = "cancelledDeletePeriod"
    FIELD_NOTIFICATION_DELAY_PERIOD = "notificationDelayPeriod"
    FIELD_CLAIMED_DELETE_PERIOD = "claimedDeletePeriod"
    FIELD_NOT_DISPENSED_DELETE_PERIOD = "notDispensedDeletePeriod"
    FIELD_RELEASE_VERSION = "releaseVersion"

    def __init__(self, log_object, internal_id):
        self.log_object = EpsLogger(log_object)
        self.internal_id = internal_id

        # Map between prescription status and method for calculating index values
        self._index_map = {}
        self._index_map[PrescriptionStatus.TO_BE_DISPENSED] = self.un_dispensed
        self._index_map[PrescriptionStatus.WITH_DISPENSER] = self.un_dispensed
        self._index_map[PrescriptionStatus.WITH_DISPENSER_ACTIVE] = self.part_dispensed
        self._index_map[PrescriptionStatus.EXPIRED] = self.expired
        self._index_map[PrescriptionStatus.CANCELLED] = self.cancelled
        self._index_map[PrescriptionStatus.DISPENSED] = self.dispensed
        self._index_map[PrescriptionStatus.NO_CLAIMED] = self.completed
        self._index_map[PrescriptionStatus.NOT_DISPENSED] = self.not_dispensed
        self._index_map[PrescriptionStatus.CLAIMED] = self.completed
        self._index_map[PrescriptionStatus.AWAITING_RELEASE_READY] = self.awaiting_nominated_release
        self._index_map[PrescriptionStatus.REPEAT_DISPENSE_FUTURE_INSTANCE] = self.un_dispensed
        self._index_map[PrescriptionStatus.FUTURE_DATED_PRESCRIPTION] = self.future_dated
        self._index_map[PrescriptionStatus.PENDING_CANCELLATION] = self.awaiting_cancellation

    def next_activity_date(self, nad_status, nad_reference):
        """
        Function takes prescriptionStatus (this will be the prescriptionStatus to be
        if the function is called during an update process)
        Function takes nad_status - a dictionary of information relevant to
        next-activity-date calculation
        Function takes nad_reference - a dictionary of global variables relevant to
        next-activity-date calculation
        Function should return [nextActivity, nextActivityDate, expiryDate]
        """
        prescription_status = nad_status[fields.FIELD_PRESCRIPTION_STATUS]

        for key in NextActivityGenerator.INPUT_BY_STATUS[prescription_status]:
            if fields.FIELD_CAPITAL_D_DATE in key:
                if nad_status[key]:
                    nad_status[key] = datetime.datetime.strptime(
                        nad_status[key], TimeFormats.STANDARD_DATE_FORMAT
                    )
                elif key not in [
                    fields.FIELD_NOMINATED_DOWNLOAD_DATE,
                    fields.FIELD_DISPENSE_WINDOW_LOW_DATE,
                ]:
                    nad_status[key] = datetime.datetime.now()

        self._calculate_expiry_date(nad_status, nad_reference)
        return_value = self._index_map[prescription_status](nad_status, nad_reference)
        return return_value

    def _calculate_expiry_date(self, nad_status, nad_reference):
        """
        Calculate the expiry date to be used in subsequent Next Activity calculations
        """
        if int(nad_status[fields.FIELD_INSTANCE_NUMBER]) > 1:
            expiry_date = (
                nad_status[fields.FIELD_PRESCRIPTION_DATE]
                + nad_reference[self.FIELD_REPEAT_DISPENSE_EXPIRY_PERIOD]
            )
        else:
            expiry_date = (
                nad_status[fields.FIELD_PRESCRIPTION_DATE]
                + nad_reference[self.FIELD_PRESCRIPTION_EXPIRY_PERIOD]
            )

        nad_status[fields.FIELD_EXPIRY_DATE] = expiry_date
        expiry_date_str = expiry_date.strftime(TimeFormats.STANDARD_DATE_FORMAT)
        nad_status[fields.FIELD_FORMATTED_EXPIRY_DATE] = expiry_date_str

    def un_dispensed(self, nad_status, _):
        """
        return [nextActivity, nextActivityDate, expiryDate] for un_dispensed prescription
        messages, covers:
        toBeDispensed
        withDispenser
        RepeatDispenseFutureInstance
        """
        next_activity = fields.NEXTACTIVITY_EXPIRE
        next_activity_date = nad_status[fields.FIELD_FORMATTED_EXPIRY_DATE]
        return [next_activity, next_activity_date, nad_status[fields.FIELD_EXPIRY_DATE]]

    def part_dispensed(self, nad_status, nad_reference):
        """
        return [nextActivity, nextActivityDate, expiryDate] for part_dispensed prescription
        messages
        """
        max_dispense_time = nad_status[fields.FIELD_LAST_DISPENSE_DATE]
        max_dispense_time += nad_reference[self.FIELD_WITH_DISPENSER_ACTIVE_EXPIRY_PERIOD]
        expiry_date = min(max_dispense_time, nad_status[fields.FIELD_EXPIRY_DATE])

        if nad_status[self.FIELD_RELEASE_VERSION] == fields.R1_VERSION:
            next_activity = fields.NEXTACTIVITY_EXPIRE
            next_activity_date = expiry_date.strftime(TimeFormats.STANDARD_DATE_FORMAT)
        else:
            if not nad_status[fields.FIELD_LAST_DISPENSE_NOTIFICATION_MSG_REF]:
                next_activity = fields.NEXTACTIVITY_EXPIRE
                next_activity_date = expiry_date.strftime(TimeFormats.STANDARD_DATE_FORMAT)
            else:
                next_activity = fields.NEXTACTIVITY_CREATENOCLAIM
                next_activity_date = max_dispense_time.strftime(TimeFormats.STANDARD_DATE_FORMAT)
        return [next_activity, next_activity_date, expiry_date]

    def expired(self, nad_status, nad_reference):
        """
        return [nextActivity, nextActivityDate, expiryDate] for expired prescription
        messages
        """
        deletion_date = (
            nad_status[fields.FIELD_COMPLETION_DATE]
            + nad_reference[self.FIELD_EXPIRED_DELETE_PERIOD]
        )
        next_activity = fields.NEXTACTIVITY_DELETE
        next_activity_date = deletion_date.strftime(TimeFormats.STANDARD_DATE_FORMAT)
        return [next_activity, next_activity_date, None]

    def cancelled(self, nad_status, nad_reference):
        """
        return [nextActivity, nextActivityDate, expiryDate] for cancelled prescription
        messages
        """
        deletion_date = (
            nad_status[fields.FIELD_COMPLETION_DATE]
            + nad_reference[self.FIELD_CANCELLED_DELETE_PERIOD]
        )
        next_activity = fields.NEXTACTIVITY_DELETE
        next_activity_date = deletion_date.strftime(TimeFormats.STANDARD_DATE_FORMAT)
        return [next_activity, next_activity_date, None]

    def dispensed(self, nad_status, nad_reference):
        """
        return [nextActivity, nextActivityDate, expiryDate] for dispensed prescription
        messages.
        Note that if a claim is not received before the notification delay period expires,
        a no claim notification is sent to the PPD.
        """
        completion_date = nad_status[fields.FIELD_COMPLETION_DATE]
        max_notification_date = (
            completion_date + nad_reference[self.FIELD_NOTIFICATION_DELAY_PERIOD]
        )
        if nad_status[self.FIELD_RELEASE_VERSION] == fields.R1_VERSION:  # noqa: SIM108
            next_activity = fields.NEXTACTIVITY_DELETE
        else:
            next_activity = fields.NEXTACTIVITY_CREATENOCLAIM
        next_activity_date = max_notification_date.strftime(TimeFormats.STANDARD_DATE_FORMAT)
        return [next_activity, next_activity_date, None]

    def completed(self, nad_status, nad_reference):
        """
        return [nextActivity, nextActivityDate, expiryDate] for completed prescription
        messages

        Note, all reference to claim sent date removed as this now only applies to already
        claimed and no-claimed prescriptions.
        """
        deletion_date = (
            nad_status[fields.FIELD_CLAIM_SENT_DATE]
            + nad_reference[self.FIELD_CLAIMED_DELETE_PERIOD]
        )
        next_activity = fields.NEXTACTIVITY_DELETE
        next_activity_date = deletion_date.strftime(TimeFormats.STANDARD_DATE_FORMAT)
        return [next_activity, next_activity_date, None]

    def not_dispensed(self, nad_status, nad_reference):
        """
        return [nextActivity, nextActivityDate, expiryDate] for not_dispensed prescription
        messages
        """
        deletion_date = (
            nad_status[fields.FIELD_COMPLETION_DATE]
            + nad_reference[self.FIELD_NOT_DISPENSED_DELETE_PERIOD]
        )
        next_activity = fields.NEXTACTIVITY_DELETE
        next_activity_date = deletion_date.strftime(TimeFormats.STANDARD_DATE_FORMAT)
        return [next_activity, next_activity_date, None]

    def awaiting_nominated_release(self, nad_status, _):
        """
        return [nextActivity, nextActivityDate, expiryDate] for awaiting_nominated_release
        prescription messages
        """
        ready_date = nad_status[fields.FIELD_DISPENSE_WINDOW_LOW_DATE]

        if nad_status[fields.FIELD_NOMINATED_DOWNLOAD_DATE]:
            ready_date = nad_status[fields.FIELD_NOMINATED_DOWNLOAD_DATE]

        ready_date_string = ready_date.strftime(TimeFormats.STANDARD_DATE_FORMAT)

        if ready_date < nad_status[fields.FIELD_EXPIRY_DATE]:
            next_activity = fields.NEXTACTIVITY_READY
            next_activity_date = ready_date_string
        else:
            next_activity = fields.NEXTACTIVITY_EXPIRE
            next_activity_date = nad_status[fields.FIELD_FORMATTED_EXPIRY_DATE]
        return [next_activity, next_activity_date, nad_status[fields.FIELD_EXPIRY_DATE]]

    def future_dated(self, nad_status, _):
        """
        return [nextActivity, nextActivityDate, expiryDate] for awaiting_nominated_release
        prescription messages
        """
        if nad_status[fields.FIELD_DISPENSE_WINDOW_LOW_DATE]:
            ready_date = max(
                nad_status[fields.FIELD_DISPENSE_WINDOW_LOW_DATE],
                nad_status[fields.FIELD_PRESCRIPTION_DATE],
            )
        else:
            ready_date = nad_status[fields.FIELD_PRESCRIPTION_DATE]

        ready_date_string = ready_date.strftime(TimeFormats.STANDARD_DATE_FORMAT)

        if nad_status[fields.FIELD_NOMINATED_DOWNLOAD_DATE]:
            ready_date = nad_status[fields.FIELD_NOMINATED_DOWNLOAD_DATE]
        if ready_date < nad_status[fields.FIELD_EXPIRY_DATE]:
            next_activity = fields.NEXTACTIVITY_READY
            next_activity_date = ready_date_string
        else:
            next_activity = fields.NEXTACTIVITY_EXPIRE
            next_activity_date = nad_status[fields.FIELD_FORMATTED_EXPIRY_DATE]
        return [next_activity, next_activity_date, nad_status[fields.FIELD_EXPIRY_DATE]]

    def awaiting_cancellation(self, nad_status, nad_reference):
        """
        return [nextActivity, nextActivityDate, expiryDate] for awaiting_cancellation
        prescription messages
        """
        deletion_date = (
            nad_status[fields.FIELD_HANDLE_TIME] + nad_reference[self.FIELD_CANCELLED_DELETE_PERIOD]
        )
        next_activity = fields.NEXTACTIVITY_DELETE
        next_activity_date = deletion_date.strftime(TimeFormats.STANDARD_DATE_FORMAT)
        return [next_activity, next_activity_date, None]



from datetime import datetime


class TimeFormats:
    STANDARD_DATE_TIME_UTC_ZONE_FORMAT = "%Y%m%d%H%M%S+0000"
    STANDARD_DATE_TIME_FORMAT = "%Y%m%d%H%M%S"
    STANDARD_DATE_TIME_LENGTH = 14
    DATE_TIME_WITHOUT_SECONDS_FORMAT = "%Y%m%d%H%M"
    STANDARD_DATE_FORMAT = "%Y%m%d"
    STANDARD_DATE_FORMAT_YEAR_MONTH = "%Y%m"
    STANDARD_DATE_FORMAT_YEAR_ONLY = "%Y"
    HL7_DATETIME_FORMAT = "%Y%m%dT%H%M%S.%f"
    SPINE_DATETIME_MS_FORMAT = "%Y%m%d%H%M%S.%f"
    SPINE_DATE_FORMAT = "%Y%m%d"
    EBXML_FORMAT = "%Y-%m-%dT%H:%M:%S"
    SMSP_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
    EXTENDED_SMSP_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"
    EXTENDED_SMSP_PLUS_Z_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


_TIMEFORMAT_LENGTH_MAP = {
    TimeFormats.STANDARD_DATE_TIME_LENGTH: TimeFormats.STANDARD_DATE_TIME_FORMAT,
    12: TimeFormats.DATE_TIME_WITHOUT_SECONDS_FORMAT,
    8: TimeFormats.STANDARD_DATE_FORMAT,
    6: TimeFormats.STANDARD_DATE_FORMAT_YEAR_MONTH,
    4: TimeFormats.STANDARD_DATE_FORMAT_YEAR_ONLY,
    22: TimeFormats.HL7_DATETIME_FORMAT,
    21: TimeFormats.SPINE_DATETIME_MS_FORMAT,
    20: TimeFormats.SMSP_FORMAT,
    23: TimeFormats.EXTENDED_SMSP_FORMAT,
    26: TimeFormats.EXTENDED_SMSP_FORMAT,
    24: TimeFormats.EXTENDED_SMSP_PLUS_Z_FORMAT,
    27: TimeFormats.EXTENDED_SMSP_PLUS_Z_FORMAT,
}


def _guess_common_datetime_format(time_string, raise_error_if_unknown=False):
    """
    Guess the date time format from the commonly used list

    Args:
        time_string (str):
            The datetime string to try determine the format of.
        raise_error_if_unknown (bool):
            Determines the action when the format cannot be determined.
            False (default) will return None, True will raise an error.
    """
    time_format = None
    if len(time_string) == 19:
        try:
            datetime.strptime(time_string, TimeFormats.EBXML_FORMAT)
            time_format = TimeFormats.EBXML_FORMAT
        except ValueError:
            time_format = TimeFormats.STANDARD_DATE_TIME_UTC_ZONE_FORMAT
    else:
        time_format = _TIMEFORMAT_LENGTH_MAP.get(len(time_string), None)

    if not time_format and raise_error_if_unknown:
        raise ValueError("Could not determine datetime format of '{}'".format(time_string))

    return time_format


def convert_spine_date(date_string, date_format=None):
    """
    Try to convert a Spine date using the passed format - if it fails - try the most
    appropriate
    """
    if date_format:
        try:
            date_object = datetime.strptime(date_string, date_format)
            return date_object
        except ValueError:
            pass

    date_format = _guess_common_datetime_format(date_string, raise_error_if_unknown=True)
    return datetime.strptime(date_string, date_format)


def time_now_as_string(date_format=TimeFormats.STANDARD_DATE_TIME_FORMAT):
    """
    Return the current date and time as a string in standard format
    """
    return now().strftime(date_format)


def now():
    """
    Utility to gets the current date and time.
    The intention is for this to be easier to replace when testing.
    :returns: a datetime representing the current date and time
    """
    return datetime.now()



import datetime
from copy import copy

from eps_spine_shared.common.prescription import fields
from eps_spine_shared.common.prescription.record import PrescriptionRecord
from eps_spine_shared.common.prescription.statuses import LineItemStatus, PrescriptionStatus


class RepeatDispenseRecord(PrescriptionRecord):
    """
    Class defined to handle repeat dispense prescriptions
    """

    def __init__(self, log_object, internal_id):
        """
        Allow the record_type attribute to be set
        """
        super(RepeatDispenseRecord, self).__init__(log_object, internal_id)
        self.record_type = "RepeatDispense"

    def create_instances(self, context, line_items):
        """
        Create all prescription instances

        Expire any lineItems that have a lower max_repeats number than the instance number
        """

        instance_snippets = {}

        range_max = int(context.maxRepeats) + 1
        future_instance_status = PrescriptionStatus.REPEAT_DISPENSE_FUTURE_INSTANCE

        for instance_number in range(1, range_max):
            instance_snippet = self.set_all_snippet_details(fields.INSTANCE_DETAILS, context)
            instance_snippet[fields.FIELD_LINE_ITEMS] = []
            for line_item in line_items:
                line_item_copy = copy(line_item)
                if int(line_item_copy[fields.FIELD_MAX_REPEATS]) < instance_number:
                    line_item_copy[fields.FIELD_STATUS] = LineItemStatus.EXPIRED
                instance_snippet[fields.FIELD_LINE_ITEMS].append(line_item_copy)

            instance_snippet[fields.FIELD_INSTANCE_NUMBER] = str(instance_number)
            if instance_number != 1:
                instance_snippet[fields.FIELD_PRESCRIPTION_STATUS] = future_instance_status
            instance_snippet[fields.FIELD_DISPENSE] = self.set_all_snippet_details(
                fields.DISPENSE_DETAILS, context
            )
            instance_snippet[fields.FIELD_CLAIM] = self.set_all_snippet_details(
                fields.CLAIM_DETAILS, context
            )
            instance_snippet[fields.FIELD_CANCELLATIONS] = []
            instance_snippet[fields.FIELD_DISPENSE_HISTORY] = {}
            instance_snippets[str(instance_number)] = instance_snippet
            instance_snippet[fields.FIELD_NEXT_ACTIVITY] = {}
            instance_snippet[fields.FIELD_NEXT_ACTIVITY][fields.FIELD_ACTIVITY] = None
            instance_snippet[fields.FIELD_NEXT_ACTIVITY][fields.FIELD_DATE] = None

        return instance_snippets

    def set_initial_prescription_status(self, handle_time):
        """
        Create the initial prescription status. For repeat dispense prescriptions, this
        needs to consider both the prescription date and the dispense window low dates.

        If either the prescriptionTime or dispenseWindowLow date is in the future then
        the prescription needs to have a Future Dated Prescription status set and can
        not yet be downloaded.
        If the prescription is not Future Dated, the default To Be Dispensed should be used.

        Note that this only applies to the first instance, the remaining instances will
        already have a Future Repeat Dispense Instance status set.

        :type handle_time: datetime.datetime
        """
        first_issue = self.get_issue(1)

        future_threshold = handle_time + datetime.timedelta(days=1)
        is_future_dated = self.time > future_threshold

        dispense_low_date = first_issue.dispense_window_low_date
        if dispense_low_date is not None and dispense_low_date > future_threshold:
            is_future_dated = True

        if is_future_dated:
            first_issue.status = PrescriptionStatus.FUTURE_DATED_PRESCRIPTION
        else:
            first_issue.status = PrescriptionStatus.TO_BE_DISPENSED

    def get_withdrawn_status(self, passed_status):
        """
        Dispense Return can only go back as far as 'with dispenser-active' for repeat dispense
        prescriptions, so convert the status for with dispenser, otherwise, return what was provided.
        """
        if passed_status == PrescriptionStatus.WITH_DISPENSER:
            return PrescriptionStatus.WITH_DISPENSER_ACTIVE
        return passed_status

    @property
    def max_repeats(self):
        """
        The maximum number of issues of this prescription.

        :rtype: int
        """
        max_repeats = self.prescription_record[fields.FIELD_PRESCRIPTION][fields.FIELD_MAX_REPEATS]
        return int(max_repeats)

    @property
    def future_issues_available(self):
        """
        Return boolean to indicate if future issues are available or not. Always False for
        Acute and Repeat Prescribe

        :rtype: bool
        """
        return self.current_issue_number < self.max_repeats



import copy
import json
import random
import sys
from datetime import datetime, timezone
from decimal import Decimal

from boto3.dynamodb.conditions import Attr, ConditionExpressionBuilder
from boto3.dynamodb.types import TypeDeserializer, TypeSerializer
from boto3.session import Session
from botocore.config import Config
from botocore.credentials import DeferredRefreshableCredentials
from botocore.exceptions import ClientError, NoCredentialsError

from eps_spine_shared.common.dynamodb_common import (
    CONDITION_EXPRESSION,
    REGION_NAME,
    SERVICE_NAME,
    Attribute,
    Key,
    ProjectedAttribute,
    SortKey,
)
from eps_spine_shared.errors import EpsNoCredentialsErrorWithRetry
from eps_spine_shared.logger import EpsLogger


class EpsDynamoDbClient:
    """
    The DynamoDB client specific to the prescriptions message store.
    """

    def __init__(
        self,
        log_object,
        aws_endpoint_url,
        table_name,
        role_arn=None,
        role_session_name=None,
        sts_endpoint_url=None,
    ):
        """
        Instantiate the DynamoDB client.
        """
        self.log_object = EpsLogger(log_object)
        self.table_name = table_name
        self.aws_endpoint_url = aws_endpoint_url
        self.role_arn = role_arn
        self.role_session_name = role_session_name
        self.sts_endpoint_url = sts_endpoint_url

        try:
            session = Session()

            if role_arn and role_session_name and sts_endpoint_url:
                credentials = DeferredRefreshableCredentials(
                    refresh_using=self._refreshed_credentials_with_retry, method="sts-assume-role"
                )
                session._session._credentials = credentials  # noqa: SLF001
            else:
                self.log_object.write_log(
                    "DDB0006",
                    None,
                    {
                        "role": role_arn,
                        "sessionName": role_session_name,
                        "endpoint": sts_endpoint_url,
                    },
                )

            resource_args = {"service_name": SERVICE_NAME, "region_name": REGION_NAME}
            if aws_endpoint_url:
                self.log_object.write_log("DDB0003", None, {"awsEndpointUrl": aws_endpoint_url})
                resource_args["endpoint_url"] = aws_endpoint_url
            else:
                self.log_object.write_log("DDB0004", None)

            self.resource = session.resource(**resource_args)
            self.table = self.resource.Table(table_name)

            self.client = session.client(**resource_args)
            self.deserialiser = TypeDeserializer()
            self.serialiser = TypeSerializer()
        except Exception as ex:
            self.log_object.write_log("DDB0000", sys.exc_info(), {"error": str(ex)})
            raise ex

        self.log_object.write_log("DDB0001", None, {"tableName": table_name})

    def _refreshed_credentials_with_retry(self, attempts=2) -> dict:
        """
        Retry _refreshed_credentials for a maximum number of attempts until credentials are returned or raise
        EpsNoCredentialsErrorWithRetry including the number of attempts.
        """
        for _ in range(attempts):
            try:
                return self._refreshed_credentials()
            except NoCredentialsError as e:
                latest_exception = e

        raise EpsNoCredentialsErrorWithRetry(attempts=attempts) from latest_exception

    def _refreshed_credentials(self) -> dict:
        """
        Refreshes the IAM credentials provided to us by STS for the duration of our session.
        This callback is invoked automatically by boto when we are past the lifetime of our
        session (uses boto3 default of refreshing 15 mins before expiry).
        DurationSeconds - duration of the role session
        RoleSessionName - becomes the User Name for subsequent api calls made with the credentials returned
        Returns:
            dict -> A dictionary containing our new set of credentials from STS as well as the
            expiration timestamp for the session.
        Adapted from nhs-aws-helpers
        """
        sts_session = Session()

        config = Config(
            connect_timeout=1,
            read_timeout=3,
            max_pool_connections=10,
            retries={"mode": "standard", "total_max_attempts": 4},
        )

        sts_client = sts_session.client(
            "sts", region_name=REGION_NAME, endpoint_url=self.sts_endpoint_url, config=config
        )

        params = {
            "RoleArn": self.role_arn,
            "RoleSessionName": self.role_session_name,
            "DurationSeconds": 3600,
        }

        # Any exceptions raised here must be caught and logged by the code using the client
        response = sts_client.assume_role(**params).get("Credentials")

        self.log_object.write_log(
            "DDB0005",
            None,
            {
                "role": self.role_arn,
                "sessionName": self.role_session_name,
                "endpoint": self.sts_endpoint_url,
            },
        )

        return {
            "access_key": response.get("AccessKeyId"),
            "secret_key": response.get("SecretAccessKey"),
            "token": response.get("SessionToken"),
            "expiry_time": response.get("Expiration").isoformat(),
        }

    def _log_item_size(self, internal_id, serialised_item):
        """
        Writes a log message including the item type and post-serialisation size in bytes.
        Bespoke sizing functions for items with compressed contents, as bytes won't serialise.
        """

        def default_size(item):
            return sys.getsizeof(json.dumps(item))

        def work_list_size(item):
            try:
                return sys.getsizeof(item["body"]["M"]["responseDetails"]["M"]["XML"]["B"])
            except KeyError:
                return default_size(item)

        def claim_size(claim):
            batch_xml = claim["body"]["M"]["Batch XML"]
            if "S" in batch_xml:
                return default_size(claim)
            batch_xml_size = sys.getsizeof(claim["body"]["M"]["Batch XML"]["B"])
            claim_deep_copy = copy.deepcopy(claim)
            del claim_deep_copy["body"]["M"]["Batch XML"]["B"]
            claim_without_batch_xml_size = sys.getsizeof(json.dumps(claim_deep_copy))
            return batch_xml_size + claim_without_batch_xml_size

        def record_size(record):
            body_size = sys.getsizeof(record["body"]["B"])
            record_deep_copy = copy.deepcopy(record)
            del record_deep_copy["body"]["B"]
            record_without_body_size = sys.getsizeof(json.dumps(record_deep_copy))
            return body_size + record_without_body_size

        def document_size(document):
            if document["pk"]["S"].startswith("Notification_"):
                return ppd_notification_size(document)

            if document["body"]["M"].get("content"):
                content_size = sys.getsizeof(document["body"]["M"]["content"]["B"])
                document_deep_copy = copy.deepcopy(document)
                del document_deep_copy["body"]["M"]["content"]["B"]
                document_without_body_size = sys.getsizeof(json.dumps(document_deep_copy))
                return content_size + document_without_body_size
            else:
                return default_size(document)

        def ppd_notification_size(document):
            document_deep_copy = copy.deepcopy(document)

            payload = None
            if document["body"]["M"]["payload"].get("B"):
                payload = document["body"]["M"]["payload"]["B"]
                del document_deep_copy["body"]["M"]["payload"]["B"]
            elif document["body"]["M"]["payload"].get("S"):
                payload = document["body"]["M"]["payload"]["S"]
                del document_deep_copy["body"]["M"]["payload"]["S"]

            payload_size = sys.getsizeof(payload)
            document_without_payload_size = sys.getsizeof(json.dumps(document_deep_copy))
            return payload_size + document_without_payload_size

        size_funcs = {
            SortKey.CLAIM.value: claim_size,
            SortKey.WORK_LIST.value: work_list_size,
            SortKey.RECORD.value: record_size,
            SortKey.DOCUMENT.value: document_size,
        }

        item_key = serialised_item["pk"]["S"]
        item_type = serialised_item["sk"]["S"]

        try:
            size = size_funcs.get(item_type, default_size)(serialised_item)
            self.log_object.write_log(
                "DDB0011",
                None,
                {
                    "itemType": item_type,
                    "table": self.table_name,
                    "key": item_key,
                    "size": size,
                    "internalID": internal_id,
                },
            )
        except Exception:  # noqa: BLE001
            self.log_object.write_log(
                "DDB0012",
                sys.exc_info(),
                {
                    "table": self.table_name,
                    "itemType": item_type,
                    "key": item_key,
                    "internalID": internal_id,
                },
            )

    def serialise_for_dynamodb(self, item):
        """
        Convert item into DynamoDB format.
        """
        return {k: self.serialiser.serialize(v) for k, v in item.items()}

    def deserialise_from_dynamodb(self, item):
        """
        Convert item from DynamoDB format.
        """
        return {k: self.deserialiser.deserialize(v) for k, v in item.items()}

    def add_condition_expression(self, put_kwargs, is_update, item):
        """
        Adds a condition expression to the put kwargs based on whether the item is being updated.
        """
        if not is_update:
            put_kwargs["ConditionExpression"] = CONDITION_EXPRESSION
        elif item[Key.SK.name] == SortKey.RECORD.value:
            put_kwargs["ExpressionAttributeNames"] = {"#currentScn": ProjectedAttribute.SCN.name}
            put_kwargs["ExpressionAttributeValues"] = self.serialise_for_dynamodb(
                {":newScn": item.get(ProjectedAttribute.SCN.name)}
            )
            put_kwargs["ConditionExpression"] = "#currentScn < :newScn"
        elif item[Key.SK.name] == SortKey.SEQUENCE_NUMBER.value:
            sequence_number = item.get(Attribute.SEQUENCE_NUMBER.name)
            if sequence_number == 1:
                return
            put_kwargs["ExpressionAttributeNames"] = {"#currentSqn": Attribute.SEQUENCE_NUMBER.name}
            put_kwargs["ExpressionAttributeValues"] = self.serialise_for_dynamodb(
                {":newSqn": sequence_number}
            )
            put_kwargs["ConditionExpression"] = "#currentSqn < :newSqn"

    def put_item(self, internal_id, item, is_update=False, log_item_size=True):
        """
        Insert an item into the configured DynamoDB table as a single put, after serialising and logging its size.
        """
        serialised_item = self.serialise_for_dynamodb(item)
        if log_item_size:
            self._log_item_size(internal_id, serialised_item)

        put_kwargs = {"TableName": self.table_name, "Item": serialised_item}
        self.add_condition_expression(put_kwargs, is_update, item)
        return self.client.put_item(**put_kwargs)

    def transact_write_items(self, internal_id, items, is_update=False, log_item_size=True):
        """
        Insert items into the configured DynamoDB table as a single transaction, after serialising and logging its size.
        """
        transact_items = []
        for item in items:
            serialised_item = self.serialise_for_dynamodb(item)
            if log_item_size:
                self._log_item_size(internal_id, serialised_item)
            transact_item = {
                "Put": {"TableName": self.table_name, "Item": self.serialise_for_dynamodb(item)}
            }
            self.add_condition_expression(transact_item["Put"], is_update, item)
            transact_items.append(transact_item)

        return self.client.transact_write_items(TransactItems=transact_items)

    def add_last_modified_to_item(self, item):
        """
        Add last modified timestamp and day to items.
        """
        dt_now = datetime.now(timezone.utc)
        last_modified_timestamp = Decimal(str(dt_now.timestamp()))
        last_modified_day = dt_now.strftime("%Y%m%d")
        partition_suffix = str(random.randint(0, 11))
        item.update(
            {
                "_lm_day": f"{last_modified_day}.{partition_suffix}",
                "_riak_lm": last_modified_timestamp,
            }
        )

    def insert_items(self, internal_id, items, is_update=False, log_item_size=True):
        """
        Perform a put_item or a transact_write_items depending on the number of items.
        """
        for item in items:
            self.add_last_modified_to_item(item)
        try:
            if len(items) == 1:
                return self.put_item(internal_id, items[0], is_update, log_item_size)
            else:
                return self.transact_write_items(internal_id, items, is_update, log_item_size)
        except ClientError as e:
            if e.response["Error"]["Code"] == "ConditionalCheckFailedException":
                pk = items[0].get(Key.PK.name)
                log_dict = {
                    "internalID": internal_id,
                    "incomingScn": items[0].get("scn", "None"),
                    "pk": pk,
                    "sk": items[0].get(Key.SK.name),
                    "table": self.table_name,
                }
                if is_update:
                    self.log_object.write_log("DDB0022", None, log_dict)
                    raise EpsDataStoreError(
                        self, pk, EpsDataStoreError.CONDITIONAL_UPDATE_FAILURE
                    ) from e
                else:
                    self.log_object.write_log("DDB0021", None, log_dict)
                    raise EpsDataStoreError(self, pk, EpsDataStoreError.DUPLICATE_ERROR) from e
            else:
                raise e

    def get_item(self, internal_id, pk, sk, expect_exists=True, expect_none=False):
        """
        Return an item from the DynamoDB table.

        expect_exists=False will not raise an error if the item does not exist.
        expect_none=True will not raise an error if the item exists but has no data.
        """
        if not pk:
            self.log_object.write_log(
                "DDB0041", None, {"key": pk, "table": self.table_name, "internalID": internal_id}
            )
            raise EpsDataStoreError(self, pk, EpsDataStoreError.ACCESS_ERROR)

        item = self.table.get_item(Key={Key.PK.name: pk, Key.SK.name: sk}).get("Item")

        self._item_checks(item, pk, expect_exists, expect_none)

        return item

    def _item_checks(self, item, key, expect_exists, expect_none):
        """
        Run standard checks on a returned item
        - Does it exist
        - Does it have data
        """
        if not expect_exists:
            return

        if item is None:
            raise EpsDataStoreError(self, key, EpsDataStoreError.MISSING_RECORD)

        if expect_none:
            return

        if item.get(ProjectedAttribute.BODY.name) is None:
            raise EpsDataStoreError(self, key, EpsDataStoreError.EMPTY_RECORD)

    def query_index(self, index_name, key_condition_expression, filter_expression):
        """
        Return the items that match the supplied expressions, for the given index.
        """
        query_args = {"KeyConditionExpression": key_condition_expression}

        if index_name:
            query_args["IndexName"] = index_name

        if filter_expression:
            query_args["FilterExpression"] = filter_expression

        items = []
        while True:
            response = self.table.query(**query_args)
            items.extend(response["Items"])
            if "LastEvaluatedKey" not in response:
                return items
            query_args["ExclusiveStartKey"] = response["LastEvaluatedKey"]

    def query_index_with_limit(
        self, index_name, key_condition_expression, filter_expression, limit
    ):
        """
        Return the items that match the supplied expressions, for the given index.
        Will return item count up to the given limit.
        """
        condition_builder = ConditionExpressionBuilder()
        key_condition_expression, condition_attributes, condition_values = (
            condition_builder.build_expression(key_condition_expression, True)
        )
        query_args = {
            "TableName": self.table_name,
            "IndexName": index_name,
            "KeyConditionExpression": key_condition_expression,
            "ExpressionAttributeNames": condition_attributes,
            "ExpressionAttributeValues": self.serialise_for_dynamodb(condition_values),
        }
        if filter_expression:
            query_args["FilterExpression"] = filter_expression
        if limit:
            query_args["Limit"] = limit

        response_iterator = self.client.get_paginator("query").paginate(**query_args)

        items = []
        for response in response_iterator:
            items.extend([self.deserialise_from_dynamodb(item) for item in response["Items"]])
            if limit and len(items) >= limit:
                items = items[:limit]
                break
        return items

    def query_index_yield(self, index_name, key_condition_expression, filter_expression=None):
        """
        Return the items that match the supplied expressions, for the given index.
        Uses yield to allow retrieval of a large number of items.
        """
        query_args = {"IndexName": index_name, "KeyConditionExpression": key_condition_expression}
        if filter_expression:
            query_args["FilterExpression"] = filter_expression

        found = True
        while found:
            response = self.table.query(**query_args)
            yield [item[Key.PK.name] for item in response["Items"]]
            if "LastEvaluatedKey" not in response:
                found = False
            else:
                query_args["ExclusiveStartKey"] = response["LastEvaluatedKey"]

    def build_filter_expression(self, filter_dict):
        """
        Build a filter expression for use in the index query.
        """
        filters = []
        for key, value in filter_dict.items():
            filters.append(Attr(key).eq(value))

        filter_expression = None
        for _filter in filters:
            filter_expression = (
                _filter if filter_expression is None else filter_expression & _filter
            )

        return filter_expression

    def delete_item(self, pk, sk):
        """
        Delete an item from the table.
        """
        key = self.serialise_for_dynamodb({Key.PK.name: pk, Key.SK.name: sk})
        self.client.delete_item(TableName=self.table_name, Key=key)


class EpsDataStoreError(Exception):
    """
    Exception to be raised when encountering issues with the DynamoDB datastore.
    """

    ACCESS_ERROR = "accessError"
    CONDITIONAL_UPDATE_FAILURE = "conditionalUpdateFailure"
    DUPLICATE_ERROR = "duplicateError"
    EMPTY_RECORD = "recordRemoved"
    MISSING_RECORD = "missingRecord"

    def __init__(self, client: EpsDynamoDbClient, key: str, error_topic: str):  # noqa: B042
        """
        The error_topic must match a topic defined as an attribute of this class (above).
        The client should have a log_object, a table_name and an aws_endpoint_url.
        """
        super(EpsDataStoreError, self).__init__()
        self.error_topic = error_topic

        log_values = {
            "awsEndpointUrl": client.aws_endpoint_url,
            "errorTopic": self.error_topic,
            "key": key,
            "tableName": client.table_name,
        }

        log_ref = "UTI0213a" if self.error_topic == self.MISSING_RECORD else "UTI0213"

        client.log_object.write_log(log_ref, None, log_values)



"""
A class to handle DynamoDB queries, such as pagination and completion flagging.
"""

from boto3.dynamodb.conditions import Attr, ConditionExpressionBuilder, Key

import eps_spine_shared.common.prescription.fields as fields
from eps_spine_shared.common.dynamodb_client import EpsDynamoDbClient
from eps_spine_shared.common.dynamodb_common import GSI, Attribute, ProjectedAttribute, RecordType
from eps_spine_shared.logger import EpsLogger


class DynamoDbQuery:
    """
    Abstraction of a dynamoDB query, handling pagination and flagging completion.
    To be used either as a generator, or calling asList.
    self.complete indicates
    """

    def __init__(
        self,
        client: EpsDynamoDbClient,
        logger: EpsLogger,
        internal_id: str,
        index: GSI,
        key_condition_expression,
        filter_expression=None,
        limit: int = None,
        descending: bool = False,
    ) -> None:
        self._client = client
        self._logger = logger
        self._internal_id = internal_id

        condition_builder = ConditionExpressionBuilder()
        key_condition_expression, condition_attributes, condition_values = (
            condition_builder.build_expression(key_condition_expression, True)
        )
        if filter_expression:
            filter_expression, filter_attributes, filter_values = (
                condition_builder.build_expression(filter_expression, False)
            )
            condition_attributes.update(filter_attributes)
            condition_values.update(filter_values)

        query_args = {
            "TableName": client.table_name,
            "IndexName": index.name,
            "KeyConditionExpression": key_condition_expression,
            "ExpressionAttributeNames": condition_attributes,
            "ExpressionAttributeValues": client.serialise_for_dynamodb(condition_values),
        }

        if filter_expression:
            query_args["FilterExpression"] = filter_expression
        if limit:
            query_args["PaginationConfig"] = {"MaxItems": limit, "PageSize": limit}
        if descending:
            query_args["ScanIndexForward"] = False

        self._pages = iter(self._client.client.get_paginator("query").paginate(**query_args))
        self._item_iterator = self._items()
        self._is_last_page = False
        self.complete = False

    def _items(self):
        """
        Yields individual deserialised items from the query, handling pagination.
        """
        for page in self._pages:
            self._logger.write_log(
                "DDB0050",
                None,
                {
                    "itemCount": len(page["Items"]),
                    "hasLastEvaluatedKey": "LastEvaluatedKey" in page,
                    "internalID": self._internal_id,
                },
            )
            items = [self._client.deserialise_from_dynamodb(item) for item in page["Items"]]
            self._is_last_page = "LastEvaluatedKey" not in page
            for item in items:
                yield item

    def __iter__(self):
        """
        Start iteration
        """
        return self

    def __next__(self):
        """
        Return the next item. If no more items, set complete flag if on last page.
        """
        try:
            return next(self._item_iterator)
        except StopIteration:
            if self._is_last_page:
                self.complete = True
            raise


class Conditions:
    """
    Wrapper for condition expressions
    """

    @staticmethod
    def nhs_number_equals(nhs_number: str):
        """
        Condition expression for nhsNumber equality
        """

        return Key(Attribute.NHS_NUMBER.name).eq(nhs_number)

    @staticmethod
    def creation_datetime_range(start: str, end: str = None):
        """
        Condition expression for creationDatetime between start and end
        """
        if not end:
            return Key(Attribute.CREATION_DATETIME.name).gte(start)

        return Key(Attribute.CREATION_DATETIME.name).between(start, end)

    @staticmethod
    def release_version_r2():
        """
        Condition expression for releaseVersion (sharded) equals R2
        """

        return Attr(ProjectedAttribute.RELEASE_VERSION.value).contains(fields.R2_VERSION)

    @staticmethod
    def next_activity_not_purged():
        """
        Condition expression for nextActivity not equal to PURGED
        """

        return ~Attr(Attribute.NEXT_ACTIVITY.name).contains(fields.NEXTACTIVITY_PURGE)

    @staticmethod
    def record_type_not_erd():
        """
        Condition expression for recordType is not repeat dispensing
        """

        return Attr(ProjectedAttribute.RECORD_TYPE.value).ne(RecordType.REPEAT_DISPENSE.value)

    @staticmethod
    def status_equals(status: str):
        """
        Condition expression for status equality
        """

        return Attr(ProjectedAttribute.STATUS.value).eq(status)



from eps_spine_shared.errors import EpsSystemError
from eps_spine_shared.logger import EpsLogger
from eps_spine_shared.nhsfundamentals.timeutilities import time_now_as_string

INDEX_NHSNUMBER_DATE = "nhsNumberDate_bin"
INDEX_NHSNUMBER_PRDATE = "nhsNumberPrescriberDate_bin"
INDEX_NHSNUMBER_PRDSDATE = "nhsNumberPrescDispDate_bin"
INDEX_NHSNUMBER_DSDATE = "nhsNumberDispenserDate_bin"
INDEX_PRESCRIBER_DATE = "prescriberDate_bin"
INDEX_PRESCRIBER_DSDATE = "prescDispDate_bin"
INDEX_PRESCRIBER_STATUS = "prescribingSiteStatus_bin"
INDEX_DISPENSER_DATE = "dispenserDate_bin"
INDEX_DISPENSER_STATUS = "dispensingSiteStatus_bin"

INDEX_NEXTACTIVITY = "nextActivityNAD_bin"
INDEX_NOMPHARM = "nomPharmStatus_bin"
INDEX_NHSNUMBER = "nhsNumber_bin"
INDEX_DELETE_DATE = "backstopdeletedate_bin"
INDEX_PRESCRIPTION_ID = "prescriptionid_bin"
INDEX_STORE_TIME_DOC_REF_TITLE = "storetimebydocreftitle_bin"

REGEX_INDICES = [
    INDEX_NHSNUMBER_DATE,
    INDEX_NHSNUMBER_PRDATE,
    INDEX_NHSNUMBER_PRDSDATE,
    INDEX_NHSNUMBER_DSDATE,
    INDEX_PRESCRIBER_DATE,
    INDEX_PRESCRIBER_DSDATE,
    INDEX_DISPENSER_DATE,
]

SEPERATOR = "|"
INDEX_DELTA = "delta_bin"


class EpsIndexFactory(object):
    """
    Factory for building index details for prescription record
    """

    def __init__(self, log_object, internal_id, test_prescribing_sites, nad_reference):
        """
        Make internal_id available for logging in indexer
        Requires nad_reference - a set of timedeltas to be used when calculating the next
        activity index
        requires test_prescribing_sites - used to differentiate for claims
        """
        self.log_object = EpsLogger(log_object)
        self.internal_id = internal_id
        self.test_prescribing_sites = test_prescribing_sites
        self.nad_reference = nad_reference

    def build_indexes(self, context):
        """
        Create the index values to be used when storing the epsRecord.  There may be
        separate index terms for each individual instance (but only unique index terms
        for the prescription should be returned).

        There are four potential indexes for the epsRecord store:
        nextActivityNAD - the next activity which is due for this prescription and the
        date which it is due (should only contain a single term)
        prescribingSiteStatus - the statuses of the prescription concatenated with the
        prescribing site (to be used in reporting and troubleshooting)
        dispensingSiteStatus - as above (not added until release has occurred)
        nomPharmStatus - as above for any nominated pharmacy (may also be used when bulk
        changes in nomination occur)
        nhsNumber - to be used when managing changes in nomination
        delta - to be used when confirming changes are synchronised between clusters
        """
        index_dict = {}
        try:
            self._add_prescibing_site_status_index(context.epsRecord, index_dict)
            self._add_dispensing_site_status_index(context.epsRecord, index_dict)
            self._add_nominated_pharmacy_status_index(context.epsRecord, index_dict)
            self._add_next_activity_next_activity_date_index(context, index_dict)
            self._add_nhs_number_index(context.epsRecord, index_dict)

            # Adding extra indexes for prescription search
            # overloading each of these indexes with Release version and prescription status in preparation for
            # Riak 1.4
            self._add_nhs_number_date_index(context.epsRecord, index_dict)
            self._add_nhs_number_prescriber_date_index(context.epsRecord, index_dict)
            self._add_nhs_number_prescriber_dispenser_date_index(context.epsRecord, index_dict)
            self._add_nhs_number_dispenser_date_index(context.epsRecord, index_dict)
            self._add_prescriber_date_index(context.epsRecord, index_dict)
            self._add_prescriber_dispenser_date_index(context.epsRecord, index_dict)
            self._add_dispenser_date_index(context.epsRecord, index_dict)
            self._add_delta_index(context.epsRecord, index_dict)
        except EpsSystemError as e:
            self.log_object.write_log(
                "EPS0124", None, {"internalID": self.internal_id, "creatingIndex": e.errorTopic}
            )
            raise EpsSystemError(EpsSystemError.MESSAGE_FAILURE) from e

        return index_dict

    def _add_nhs_number_date_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        nhs_number = eps_record.return_nhs_number()
        prescription_time = eps_record.return_prescription_time()
        nhs_number_date_bin = nhs_number + SEPERATOR + prescription_time
        index_dict[INDEX_NHSNUMBER_DATE] = eps_record.add_release_and_status(nhs_number_date_bin)

    def _add_nhs_number_prescriber_date_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        nhs_number = eps_record.return_nhs_number()
        prescriber = eps_record.return_prescribing_organisation()
        prescription_time = eps_record.return_prescription_time()
        index = nhs_number + SEPERATOR + prescriber + SEPERATOR + prescription_time
        new_indexes = eps_record.add_release_and_status(index)
        index_dict[INDEX_NHSNUMBER_PRDATE] = new_indexes

    def _add_nhs_number_prescriber_dispenser_date_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        result_list = eps_record.return_nhs_number_prescriber_dispenser_date_index()
        [success, nhs_number_presc_disp_date_bin] = result_list
        if not success:
            raise EpsSystemError(INDEX_NHSNUMBER_PRDSDATE)
        if nhs_number_presc_disp_date_bin:
            new_indexes = eps_record.add_release_and_status(nhs_number_presc_disp_date_bin, False)
            index_dict[INDEX_NHSNUMBER_PRDSDATE] = new_indexes

    def _add_prescriber_date_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        prescriber = eps_record.return_prescribing_organisation()
        prescription_time = eps_record.return_prescription_time()
        prescriber_date_bin = prescriber + SEPERATOR + prescription_time
        index_dict[INDEX_PRESCRIBER_DATE] = eps_record.add_release_and_status(prescriber_date_bin)

    def _add_nhs_number_dispenser_date_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        result_list = eps_record.return_nhs_number_dispenser_date_index()
        [success, nhs_number_dispenser_date_bin] = result_list
        if not success:
            raise EpsSystemError(INDEX_NHSNUMBER_DSDATE)
        if nhs_number_dispenser_date_bin:
            new_indexes = eps_record.add_release_and_status(nhs_number_dispenser_date_bin, False)
            index_dict[INDEX_NHSNUMBER_DSDATE] = new_indexes

    def _add_prescriber_dispenser_date_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        result_list = eps_record.return_prescriber_dispenser_date_index()
        [success, presc_disp_dates] = result_list
        if not success:
            raise EpsSystemError(INDEX_PRESCRIBER_DSDATE)
        if presc_disp_dates:
            new_indexes = eps_record.add_release_and_status(presc_disp_dates, False)
            index_dict[INDEX_PRESCRIBER_DSDATE] = new_indexes

    def _add_dispenser_date_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        result_list = eps_record.return_dispenser_date_index()
        [success, disp_dates] = result_list
        if not success:
            raise EpsSystemError(INDEX_DISPENSER_DATE)
        if disp_dates:
            new_indexes = eps_record.add_release_and_status(disp_dates, False)
            index_dict[INDEX_DISPENSER_DATE] = new_indexes

    def _add_next_activity_next_activity_date_index(self, context, index_dict):
        """
        See build_indexes
        """
        result_list = context.epsRecord.return_next_activity_index(
            self.test_prescribing_sites, self.nad_reference, context
        )

        [next_activity, next_activity_date] = result_list
        next_activity_nad_bin = (
            f"{next_activity}_{next_activity_date}"
            if next_activity_date and next_activity
            else next_activity
        )
        index_dict[INDEX_NEXTACTIVITY] = [next_activity_nad_bin]

    def _add_prescibing_site_status_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        result_list = eps_record.return_presc_site_status_index()
        [success, presc_site, prescription_status] = result_list
        if not success:
            raise EpsSystemError(INDEX_PRESCRIBER_STATUS)
        index_dict[INDEX_PRESCRIBER_STATUS] = []
        for status in prescription_status:
            index_dict[INDEX_PRESCRIBER_STATUS].append(presc_site + "_" + status)

    def _add_dispensing_site_status_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        result_list = eps_record.return_disp_site_status_index()
        [success, disp_site_statuses] = result_list
        if not success:
            raise EpsSystemError(INDEX_DISPENSER_STATUS)
        index_dict[INDEX_DISPENSER_STATUS] = list(disp_site_statuses)

    def _add_nominated_pharmacy_status_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        [nom_pharmacy, prescription_status] = eps_record.return_nom_pharm_status_index()

        if nom_pharmacy:
            index_dict[INDEX_NOMPHARM] = []
            for status in prescription_status:
                index_dict[INDEX_NOMPHARM].append(nom_pharmacy + "_" + status)

            self.log_object.write_log(
                "EPS0617",
                None,
                {
                    "internalID": self.internal_id,
                    "nomPharmacy": nom_pharmacy,
                    "indexes": index_dict[INDEX_NOMPHARM],
                },
            )
        else:
            self.log_object.write_log("EPS0618", None, {"internalID": self.internal_id})

    def _add_nhs_number_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        nhs_number = eps_record.return_nhs_number()
        index_dict[INDEX_NHSNUMBER] = [nhs_number]

    def _add_delta_index(self, eps_record, index_dict):
        """
        See build_indexes
        """
        index_dict[INDEX_DELTA] = [time_now_as_string() + SEPERATOR + str(eps_record.get_scn())]



import base64
import copy
import functools
import sys
import time
import zlib
from datetime import datetime, timedelta, timezone
from random import randint
from typing import List, Tuple

import simplejson
from boto3.dynamodb.types import Binary
from dateutil.relativedelta import relativedelta

from eps_spine_shared.common import indexes
from eps_spine_shared.common.dynamodb_client import EpsDataStoreError, EpsDynamoDbClient
from eps_spine_shared.common.dynamodb_common import (
    GSI,
    NEXT_ACTIVITY_DATE_PARTITIONS,
    Attribute,
    Key,
    ProjectedAttribute,
    SortKey,
    determine_release_version,
    prescription_id_without_check_digit,
    replace_decimals,
)
from eps_spine_shared.common.dynamodb_index import EpsDynamoDbIndex, PrescriptionStatus
from eps_spine_shared.common.dynamodb_query import Conditions, DynamoDbQuery
from eps_spine_shared.logger import EpsLogger
from eps_spine_shared.nhsfundamentals.timeutilities import (
    TimeFormats,
    convert_spine_date,
    time_now_as_string,
)


def timer(func):
    """
    Decorator to be used to time methods.
    """

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        self = args[0]
        internal_id = args[1]
        start_time = time.perf_counter()
        value = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time_ms = (end_time - start_time) * 1000
        run_time_ms = float(f"{run_time_ms:.2f}")
        self.log_object.write_log(
            "DDB0002",
            None,
            {
                "cls": type(self).__name__,
                "func": func.__name__,
                "duration": run_time_ms,
                "internalID": internal_id,
            },
        )
        return value

    return wrapper_timer


class EpsDynamoDbDataStore:
    """
    The prescriptions message store specific DynamoDB client.
    """

    SEPARATOR = "#"
    CLAIM_SEQUENCE_NUMBER_KEY = "claimSequenceNumber"
    NWSSP_CLAIM_SEQUENCE_NUMBER_KEY = "claimSequenceNumberNwssp"
    INDEX_CLAIMID = "claimid_bin"
    INDEX_CLAIMHANDLETIME = "claimhandletime_bin"
    INDEX_CLAIM_SEQNUMBER = "seqnum_bin"
    INDEX_CLAIM_SEQNUMBER_NWSSP = "nwsspseqnum_bin"
    INDEX_SCN = "delta_bin"
    INDEX_WORKLISTDATE = "workListDate_bin"
    NOTIFICATION_PREFIX = "Notification_"
    STORE_TIME_DOC_REF_TITLE_PREFIX = "NominatedReleaseRequestMsgRef"
    DEFAULT_EXPIRY_DAYS = 56
    MAX_NEXT_ACTIVITY_DATE = "99991231"

    def __init__(self, log_object, system_config):
        """
        Instantiate the DynamoDB client.
        """
        self.log_object = EpsLogger(log_object)
        self.client = EpsDynamoDbClient(
            log_object,
            system_config["ddb aws endpoint url"],
            system_config["datastore table name"],
            system_config["datastore role arn"],
            system_config["process name"],
            system_config["sts endpoint url"],
        )
        self.indexes = EpsDynamoDbIndex(log_object, self.client)

    def testConnection(self):
        """
        Placeholder test connection, returns constant value
        """
        return True

    def base64_decode_document_content(self, internal_id, document):
        """
        base64 decode document content in order to store as binary type in DynamoDB.
        """
        if content := document.get("content"):
            try:
                decoded = base64.b64decode(document["content"].encode("utf-8"))
                if base64.b64encode(decoded).decode("utf-8") == content:
                    document["content"] = decoded
                else:
                    raise ValueError("Document content not b64 encoded")
            except Exception as e:  # noqa: BLE001
                self.log_object.write_log(
                    "DDB0031", sys.exc_info(), {"error": str(e), "internalID": internal_id}
                )
                raise e

    def get_expire_at(self, delta, from_datetime=None):
        """
        Returns an int timestamp to be used as an expireAt attribute.
        This will determine when the item is deleted from the table.
        """
        if not from_datetime:
            from_datetime = datetime.now(timezone.utc)

        if not from_datetime.tzinfo:
            from_datetime = datetime.combine(
                from_datetime.date(), from_datetime.time(), timezone.utc
            )

        return int((from_datetime + delta).timestamp())

    def calculate_record_expire_at(
        self, next_activity, next_activity_date_str, creation_datetime_string
    ):
        """
        If the record next activity is delete or purge, use the nextActivity and nextActivityDate
        to calculate its expireAt (ttl) value, otherwise fall-back to the default of 18 months.
        """
        creation_datetime = convert_spine_date(
            creation_datetime_string, TimeFormats.STANDARD_DATE_TIME_FORMAT
        )
        creation_datetime_utc = datetime.combine(
            creation_datetime.date(), creation_datetime.time(), timezone.utc
        )
        default_expire_at = self.get_expire_at(relativedelta(months=18), creation_datetime_utc)

        if (
            next_activity.lower() not in ["delete", "purge"]
            or not next_activity_date_str
            or next_activity_date_str == self.MAX_NEXT_ACTIVITY_DATE
        ):
            return default_expire_at

        delta = relativedelta() if next_activity.lower() == "purge" else relativedelta(months=12)

        next_activity_datetime = convert_spine_date(
            next_activity_date_str, TimeFormats.STANDARD_DATE_FORMAT
        )
        next_activity_datetime_utc = datetime.combine(
            next_activity_datetime.date(), next_activity_datetime.time(), timezone.utc
        )

        next_activity_expire_at = self.get_expire_at(delta, next_activity_datetime_utc)

        return min(next_activity_expire_at, default_expire_at)

    def build_document(self, internal_id, document, index):
        """
        Build EPS Document object to be inserted into DynamoDB.
        """
        document_copy = copy.deepcopy(document)
        self.base64_decode_document_content(internal_id, document_copy)

        default_expire_at = self.get_expire_at(relativedelta(months=18))

        item = {
            Key.SK.name: SortKey.DOCUMENT.value,
            ProjectedAttribute.INDEXES.name: self.convert_index_keys_to_lower_case(index),
            ProjectedAttribute.BODY.name: document_copy,
            ProjectedAttribute.EXPIRE_AT.name: default_expire_at,
        }

        if index:
            doc_ref_title, store_time = index[indexes.INDEX_STORE_TIME_DOC_REF_TITLE][0].split("_")
            item[Attribute.DOC_REF_TITLE.name] = doc_ref_title

            if doc_ref_title == "ClaimNotification":
                item[Attribute.CLAIM_NOTIFICATION_STORE_DATE.name] = store_time[:8]

            item[Attribute.STORE_TIME.name] = store_time

            delete_date = index[indexes.INDEX_DELETE_DATE][0]
            delete_date_time = datetime.strptime(delete_date, TimeFormats.STANDARD_DATE_FORMAT)
            item[ProjectedAttribute.EXPIRE_AT.name] = int(delete_date_time.timestamp())

        return item

    @timer
    def insert_eps_document_object(self, internal_id, document_key, document, index=None):
        """
        Insert EPS Document object into the configured table.
        """
        item = self.build_document(internal_id, document, index)
        item[Key.PK.name] = document_key
        return self.client.insert_items(internal_id, [item], True)

    def convert_index_keys_to_lower_case(self, index):
        """
        Convert all keys in an index dict to lower case.
        """
        if not isinstance(index, dict):
            return index
        return {key.lower(): index[key] for key in index}

    def parse_next_activity_nad(self, indexes):
        """
        Split nextActivityNAD string into sharded next activity and its date.
        """
        next_activity_nad = indexes["nextActivityNAD_bin"][0]
        next_activity_nad_split = next_activity_nad.split("_")

        next_activity = next_activity_nad_split[0]
        next_activity_date_str = (
            next_activity_nad_split[1] if len(next_activity_nad_split) == 2 else None
        )

        shard = randint(1, NEXT_ACTIVITY_DATE_PARTITIONS)

        return next_activity, shard, next_activity_date_str

    def build_record(self, prescription_id, record, record_type, indexes):
        """
        Build EPS Record object to be inserted into DynamoDB.
        """
        record_key = prescription_id_without_check_digit(prescription_id)

        if not indexes:
            indexes = record["indexes"]
        instances = record["instances"].values()

        next_activity, shard, next_activity_date_str = self.parse_next_activity_nad(indexes)

        scn = record["SCN"]

        compressed_record = zlib.compress(simplejson.dumps(record).encode("utf-8"))

        creation_datetime_string = record["prescription"]["prescriptionTime"]

        expire_at = self.calculate_record_expire_at(
            next_activity, next_activity_date_str, creation_datetime_string
        )

        item = {
            Key.PK.name: record_key,
            Key.SK.name: SortKey.RECORD.value,
            ProjectedAttribute.BODY.name: compressed_record,
            Attribute.NEXT_ACTIVITY.name: f"{next_activity}.{shard}",
            ProjectedAttribute.SCN.name: scn,
            ProjectedAttribute.INDEXES.name: self.convert_index_keys_to_lower_case(indexes),
            ProjectedAttribute.EXPIRE_AT.name: expire_at,
        }
        if next_activity_date_str:
            item[Attribute.NEXT_ACTIVITY_DATE.name] = next_activity_date_str

        if next_activity.lower() == "purge":
            return item

        nhs_number = record["patient"]["nhsNumber"]
        prescriber_org = record["prescription"]["prescribingOrganization"]

        statuses = list(set([instance["prescriptionStatus"] for instance in instances]))
        is_ready = PrescriptionStatus.TO_BE_DISPENSED in statuses
        if PrescriptionStatus.TO_BE_DISPENSED in statuses:
            statuses.remove(PrescriptionStatus.TO_BE_DISPENSED)
            statuses.insert(0, PrescriptionStatus.TO_BE_DISPENSED)
        status = self.SEPARATOR.join(statuses)

        dispenser_orgs = []
        for instance in instances:
            org = instance.get("dispense", {}).get("dispensingOrganization")
            if org:
                dispenser_orgs.append(org)
        dispenser_org = self.SEPARATOR.join(set(dispenser_orgs))

        nominated_pharmacy = record.get("nomination", {}).get("nominatedPerformer")

        item_update = {
            Attribute.CREATION_DATETIME.name: creation_datetime_string,
            Attribute.NHS_NUMBER.name: nhs_number,
            Attribute.PRESCRIBER_ORG.name: prescriber_org,
            ProjectedAttribute.STATUS.name: status,
            Attribute.IS_READY.name: int(is_ready),
        }
        if dispenser_org:
            item[Attribute.DISPENSER_ORG.name] = dispenser_org
        if nominated_pharmacy:
            item[Attribute.NOMINATED_PHARMACY.name] = nominated_pharmacy
            if not dispenser_org:
                item[Attribute.DISPENSER_ORG.name] = nominated_pharmacy
        if record_type:
            item[ProjectedAttribute.RECORD_TYPE.name] = record_type
        item[ProjectedAttribute.RELEASE_VERSION.name] = determine_release_version(prescription_id)

        item.update(item_update)
        return item

    @timer
    def insert_eps_record_object(
        self, internal_id, prescription_id, record, index=None, record_type=None, is_update=False
    ):
        """
        Insert EPS Record object into the configured table.
        """
        item = self.build_record(prescription_id, record, record_type, index)

        return self.client.insert_items(internal_id, [item], is_update)

    @timer
    def insert_eps_work_list(self, internal_id, message_id, work_list, index=None):
        """
        Insert EPS WorkList object into the configured table.
        """
        work_list_indexes = {self.INDEX_WORKLISTDATE: [time_now_as_string()]}
        if index:
            work_list_indexes = index

        expire_at = self.get_expire_at(timedelta(days=self.DEFAULT_EXPIRY_DAYS))
        item = {
            Key.PK.name: message_id,
            Key.SK.name: SortKey.WORK_LIST.value,
            ProjectedAttribute.EXPIRE_AT.name: expire_at,
            ProjectedAttribute.BODY.name: self.compress_work_list_xml(internal_id, work_list),
            ProjectedAttribute.INDEXES.name: self.convert_index_keys_to_lower_case(
                work_list_indexes
            ),
        }
        return self.client.insert_items(internal_id, [item], True)

    @timer
    def is_record_present(self, internal_id, prescription_id) -> bool:
        """
        Returns a boolean indicating the presence of a record.
        """
        record_key = prescription_id_without_check_digit(prescription_id)
        record = self.client.get_item(
            internal_id, record_key, SortKey.RECORD.value, expect_exists=False
        )
        return True if record else False

    @timer
    def return_terms_by_nhs_number_date(self, internal_id, range_start, range_end, term_regex=None):
        """
        Return the epsRecord terms which match the supplied range and regex for the nhsNumberDate index.
        """
        return self.return_terms_by_index_date(
            internal_id, indexes.INDEX_NHSNUMBER_DATE, range_start, range_end, term_regex
        )

    @timer
    def return_terms_by_index_date(
        self, _internal_id, index, range_start, range_end=None, term_regex=None
    ):
        """
        Return the epsRecord terms which match the supplied range and regex for the supplied index.
        """
        index_map = {
            indexes.INDEX_NHSNUMBER_PRDSDATE: self.indexes.nhs_number_presc_disp_date,
            indexes.INDEX_NHSNUMBER_PRDATE: self.indexes.nhs_number_presc_date,
            indexes.INDEX_NHSNUMBER_DSDATE: self.indexes.nhs_number_disp_date,
            indexes.INDEX_NHSNUMBER_DATE: self.indexes.nhs_number_date,
            indexes.INDEX_PRESCRIBER_DSDATE: self.indexes.presc_disp_date,
            indexes.INDEX_PRESCRIBER_DATE: self.indexes.presc_date,
            indexes.INDEX_DISPENSER_DATE: self.indexes.disp_date,
            indexes.INDEX_NOMPHARM: self.indexes.nom_pharm_status,
        }
        return index_map[index](range_start, range_end, term_regex)

    @timer
    def return_terms_by_nhs_number(self, _internal_id, nhs_number):
        """
        Return the epsRecord terms which match the supplied NHS number.
        """
        return self.indexes.query_nhs_number_date(indexes.INDEX_NHSNUMBER, nhs_number)

    @timer
    def return_pids_for_nomination_change(self, internal_id, nhs_number):
        """
        Return the epsRecord list which match the supplied NHS number.
        """
        pid_list = self.return_terms_by_nhs_number(internal_id, nhs_number)

        prescriptions = []

        for pid in pid_list:
            prescriptions.append(pid[1])

        return prescriptions

    def get_nominated_pharmacy_records(self, nominated_pharmacy, batch_size, internal_id):
        """
        Run an index query to get the to-be-dispensed prescriptions for this nominated pharmacy.
        """
        key_list = self.get_nom_pharm_records_unfiltered(internal_id, nominated_pharmacy)
        discarded_key_count = max((len(key_list) - int(batch_size)), 0)
        key_list = key_list[:batch_size]
        return [key_list, discarded_key_count]

    @timer
    def get_nom_pharm_records_unfiltered(self, _internal_id, nominated_pharmacy, limit=None):
        """
        Query the nomPharmStatus index to get the unfiltered, to-be-dispensed prescriptions for the given pharmacy.
        """
        return self.indexes.query_nom_pharm_status(nominated_pharmacy, limit=limit)

    @timer
    def return_record_for_process(self, internal_id, prescription_id, expect_exists=True):
        """
        Look for and return an epsRecord object.
        """
        record_key = prescription_id_without_check_digit(prescription_id)
        item = self.client.get_item(
            internal_id, record_key, SortKey.RECORD.value, expect_exists=expect_exists
        )
        if not item:
            return {}
        body = item.get(ProjectedAttribute.BODY.name)
        if body and not isinstance(body, dict):
            body = simplejson.loads(zlib.decompress(bytes(body)))

        return self._build_record_to_return(item, body)

    def _build_record_to_return(self, item, body):
        """
        Create the record in the format expected by the calling code.
        """
        replace_decimals(body)

        record = {"value": body, "vectorClock": "vc"}

        if record_type := item.get("recordType"):
            record["recordType"] = record_type

        sharded_release_version = item.get(
            "releaseVersion", determine_release_version(item.get(Key.PK.name))
        )
        record["releaseVersion"] = sharded_release_version.split(".")[0]

        return record

    def base64_encode_document_content(self, internal_id, document_body):
        """
        base64 encode document content and convert to string, to align with return type of original datastore.
        """
        if document_body and not isinstance(document_body.get("content"), str):
            try:
                document_body["content"] = base64.b64encode(bytes(document_body["content"])).decode(
                    "utf-8"
                )
            except Exception as e:  # noqa: BLE001
                self.log_object.write_log(
                    "DDB0032", sys.exc_info(), {"error": str(e), "internalID": internal_id}
                )
                raise e

    @timer
    def return_document_for_process(self, internal_id, document_key, expect_exists=True):
        """
        Look for and return an epsDocument object.
        """
        item = self.client.get_item(
            internal_id,
            document_key,
            SortKey.DOCUMENT.value,
            expect_none=True,
            expect_exists=expect_exists,
        )
        if not item:
            return {}

        body = item.get(ProjectedAttribute.BODY.name)
        replace_decimals(body)

        if item.get(Attribute.DOC_REF_TITLE.name, "").lower() != "claimnotification":
            self.base64_encode_document_content(internal_id, body)
        elif isinstance(body.get("payload"), Binary):
            body["payload"] = body["payload"].value.decode("utf-8")

        return body

    @timer
    def return_record_for_update(self, internal_id, prescription_id):
        """
        Look for and return an epsRecord object,
        but with dataObject on self so that an update can be applied.
        """
        record_key = prescription_id_without_check_digit(prescription_id)
        item = self.client.get_item(internal_id, record_key, SortKey.RECORD.value)
        body = item.get(ProjectedAttribute.BODY.name)
        if body and not isinstance(body, dict):
            body = simplejson.loads(zlib.decompress(bytes(body)))

        self.dataObject = body
        return self._build_record_to_return(item, body)

    def get_prescription_record_data(self, internal_id, prescription_id, expect_exists=True):
        """
        Gets the prescription record from the data store and return just the data.
        :expect_exists defaulted to True. Thus we expect the key should already exist, if
        no matches are found DDB will throw a EpsDataStoreError (Missing Record).
        """
        record_key = prescription_id_without_check_digit(prescription_id)
        data_object = self.client.get_item(
            internal_id, record_key, SortKey.RECORD.value, expect_exists=expect_exists
        )

        if data_object is None:
            return None

        return data_object

    @timer
    def get_work_list(self, internal_id, message_id):
        """
        Look for and return a workList object.
        """
        item = self.client.get_item(
            internal_id, message_id, SortKey.WORK_LIST.value, expect_exists=False, expect_none=True
        )
        if item is None:
            return None

        if body := item.get(ProjectedAttribute.BODY.name):
            replace_decimals(body)
            self.decompress_work_list_xml(internal_id, body)
        return body

    @timer
    def compress_work_list_xml(self, _internal_id, work_list):
        """
        Compresses the XML contained in the work list, if present. Maintains original responseDetails on context.
        """
        work_list_deep_copy = copy.deepcopy(work_list)
        xml_bytes = work_list_deep_copy.get("responseDetails", {}).get("XML")

        if xml_bytes:
            if isinstance(xml_bytes, str):
                xml_bytes = xml_bytes.encode("utf-8")
            compressed_xml = zlib.compress(xml_bytes)
            work_list_deep_copy["responseDetails"]["XML"] = compressed_xml
        return work_list_deep_copy

    @timer
    def decompress_work_list_xml(self, _internal_id, body):
        """
        Decompresses the XML contained in the work list, if present.
        """
        compressed_xml = body.get("responseDetails", {}).get("XML")

        if compressed_xml:
            decompressed_xml = zlib.decompress(bytes(compressed_xml))
            body["responseDetails"]["XML"] = decompressed_xml

    def _fetch_next_sequence_number(self, internal_id, key, max_sequence_number, read_only=False):
        """
        Fetch the next sequence number from a given key.
        """
        item = self.client.get_item(
            internal_id, key, SortKey.SEQUENCE_NUMBER.value, expect_exists=False
        )
        is_update = True
        if not item:
            item = {
                Key.PK.name: key,
                Key.SK.name: SortKey.SEQUENCE_NUMBER.value,
                Attribute.SEQUENCE_NUMBER.name: 1,
            }
            is_update = False
        else:
            replace_decimals(item)
            sequence_number = item[Attribute.SEQUENCE_NUMBER.name]
            item[Attribute.SEQUENCE_NUMBER.name] = (
                sequence_number + 1 if sequence_number < max_sequence_number else 1
            )

        if not read_only:
            tries = 0
            while True:
                try:
                    self.client.insert_items(internal_id, [item], is_update, False)
                    break
                except EpsDataStoreError as e:
                    if e.error_topic == EpsDataStoreError.CONDITIONAL_UPDATE_FAILURE and tries < 25:
                        sequence_number = item[Attribute.SEQUENCE_NUMBER.name]
                        item[Attribute.SEQUENCE_NUMBER.name] = (
                            sequence_number + 1 if sequence_number < max_sequence_number else 1
                        )
                        tries += 1
                    else:
                        raise

        return item[Attribute.SEQUENCE_NUMBER.name]

    @timer
    def fetch_next_sequence_number(self, internal_id, max_sequence_number, read_only=False):
        """
        Fetch the next sequence number for a batch claim message.
        ONLY SINGLETON WORKER PROCESSES SHOULD CALL THIS - IT IS NOT AN ATOMIC ACTION.
        """
        return self._fetch_next_sequence_number(
            internal_id, self.CLAIM_SEQUENCE_NUMBER_KEY, max_sequence_number, read_only
        )

    @timer
    def fetch_next_sequence_number_nwssp(self, internal_id, max_sequence_number, read_only=False):
        """
        Fetch the next sequence number for a welsh batch claim message

        ONLY SINGLETON WORKER PROCESSES SHOULD CALL THIS - IT IS NOT AN ATOMIC ACTION
        """
        return self._fetch_next_sequence_number(
            internal_id, self.NWSSP_CLAIM_SEQUENCE_NUMBER_KEY, max_sequence_number, read_only
        )

    @timer
    def store_batch_claim(self, internal_id, batch_claim_original):
        """
        batchClaims need to be stored by their GUIDs with a claims sort key.
        They also require an index value for each claimID in the batch.
        A further index value is added with sequence number, for batch resend functionality.
        """
        batch_claim = copy.deepcopy(batch_claim_original)
        key = batch_claim["Batch GUID"]

        claim_id_index_terms = batch_claim["Claim ID List"]
        handle_time_index_term = batch_claim["Handle Time"]
        sequence_number = batch_claim["Sequence Number"]
        index_scn_value = f"{time_now_as_string()}|{sequence_number}"

        nwssp = "Nwssp Sequence Number" in batch_claim
        nwssp_sequence_number = batch_claim.get("Nwssp Sequence Number")
        expire_at = self.get_expire_at(timedelta(days=self.DEFAULT_EXPIRY_DAYS))

        indexes = {
            self.INDEX_CLAIMID: claim_id_index_terms,
            self.INDEX_CLAIMHANDLETIME: [handle_time_index_term],
            self.INDEX_CLAIM_SEQNUMBER: [sequence_number],
            self.INDEX_SCN: [index_scn_value],
        }
        if nwssp:
            indexes[self.INDEX_CLAIM_SEQNUMBER_NWSSP] = [nwssp_sequence_number]

        if batch_claim.get("Claim Metadata") and not batch_claim.get("Backward Incompatible"):
            batch_claim["Batch XML"] = ""

        item = {
            Key.PK.name: key,
            Key.SK.name: SortKey.CLAIM.value,
            ProjectedAttribute.BODY.name: batch_claim,
            ProjectedAttribute.EXPIRE_AT.name: expire_at,
            ProjectedAttribute.CLAIM_IDS.name: claim_id_index_terms,
            ProjectedAttribute.INDEXES.name: self.convert_index_keys_to_lower_case(indexes),
            Attribute.BATCH_CLAIM_ID.name: key,
        }
        if nwssp:
            item[Attribute.SEQUENCE_NUMBER_NWSSP.name] = nwssp_sequence_number
        else:
            item[Attribute.SEQUENCE_NUMBER.name] = sequence_number

        try:
            self.client.insert_items(internal_id, [item], True)
        except Exception:  # noqa: BLE001
            self.log_object.write_log("EPS0279", sys.exc_info(), {"internalID": key})
            return False
        return True

    def fetch_batch_claim(self, internal_id, batch_claim_id):
        """
        Retrieves the batch claim and returns the batch message for the calling application to handle.
        """
        item = self.client.get_item(
            internal_id, batch_claim_id, SortKey.CLAIM.value, expect_exists=False
        )
        if not item:
            return {}

        body = item.get(ProjectedAttribute.BODY.name)
        replace_decimals(body)
        batch_xml = body["Batch XML"]

        if not isinstance(batch_xml, str):
            try:
                body["Batch XML"] = bytes(batch_xml).decode("utf-8")
            except Exception as e:  # noqa: BLE001
                self.log_object.write_log(
                    "DDB0033", sys.exc_info(), {"error": str(e), "internalID": internal_id}
                )
                raise e

        return body

    @timer
    def delete_claim_notification(self, internal_id, claim_id):
        """
        Delete the claim notification document from the table, and return True if the deletion was successful.
        """
        try:
            self.client.delete_item(
                self.NOTIFICATION_PREFIX + str(claim_id), SortKey.DOCUMENT.value
            )
        except Exception:  # noqa: BLE001
            self.log_object.write_log(
                "EPS0289", sys.exc_info(), {"claimID": claim_id, "internalID": internal_id}
            )
            return False
        return True

    @timer
    def delete_document(self, internal_id, document_key, delete_notification=False):
        """
        Delete a document from the table. Return a boolean indicator of success.
        """
        if (
            str(document_key).lower().startswith(self.NOTIFICATION_PREFIX.lower())
            and not delete_notification
        ):
            return True

        item = self.client.get_item(
            internal_id, document_key, SortKey.DOCUMENT.value, expect_exists=False
        )

        if not item:
            self.log_object.write_log(
                "EPS0601b", None, {"documentRef": document_key, "internalID": internal_id}
            )
            return False

        self.log_object.write_log(
            "EPS0601", None, {"documentRef": document_key, "internalID": internal_id}
        )
        self.client.delete_item(document_key, SortKey.DOCUMENT.value)
        return True

    @timer
    def delete_record(self, internal_id, record_key):
        """
        Delete a record from the table.
        """
        self.log_object.write_log(
            "EPS0602", None, {"recordRef": record_key, "internalID": internal_id}
        )
        self.client.delete_item(record_key, SortKey.RECORD.value)

    @timer
    def return_pids_due_for_next_activity(
        self, _internal_id, next_activity_start, next_activity_end, shard=None
    ):
        """
        Returns all the epsRecord keys for prescriptions whose nextActivity is the same as that provided,
        and whose next activity date is within the date range provided.
        """
        return self.indexes.query_next_activity_date(next_activity_start, next_activity_end, shard)

    @timer
    def return_prescription_ids_for_nom_pharm(self, _internal_id, nominated_pharmacy_index_term):
        """
        Returns the epsRecord keys relating to the given nominated pharmacy term.
        """
        ods_code = nominated_pharmacy_index_term.split("_")[0]
        return self.indexes.query_nom_pharm_status(ods_code)

    @timer
    def return_claim_notification_ids_between_store_dates(self, internal_id, start_date, end_date):
        """
        Returns all the epsDocument keys for claim notification documents whose store dates are in the given window.
        """
        return self.indexes.query_claim_notification_store_time(internal_id, start_date, end_date)

    @timer
    def get_all_pids_by_nominated_pharmacy(self, _internal_id, nominated_pharmacy):
        """
        Run an index query to get all prescriptions for this nominated pharmacy.
        """
        return self.indexes.query_nom_pharm_status(nominated_pharmacy, True)

    @timer
    def check_item_exists(self, internal_id, pk, sk, expect_exists) -> bool:
        """
        Returns False as covered by condition expression.
        """
        item = self.client.get_item(internal_id, pk, sk, expect_exists)
        if item:
            return True
        return False

    def find_batch_claim_from_seq_number(self, sequence_number, nwssp=False):
        """
        Run a query against the sequence number index looking for the
        batch GUID (key) on the basis of sequence number.
        """
        return self.indexes.query_batch_claim_id_sequence_number(sequence_number, nwssp)

    @timer
    def return_pfp_pids_for_nhs_number(
        self, internal_id, nhs_number, start_date, end_date, limit
    ) -> Tuple[bool, List[str]]:
        """
        Returns a list of prescription IDs against a given NHS number for PfP.
        Also returns a boolean indicating if there are more results available.
        """
        key_conditions = Conditions.nhs_number_equals(
            nhs_number
        ) & Conditions.creation_datetime_range(start_date, end_date)

        filter_expressions = (
            Conditions.release_version_r2()
            & Conditions.next_activity_not_purged()
            & Conditions.record_type_not_erd()
        )

        desired_statuses = [
            PrescriptionStatus.TO_BE_DISPENSED,
            PrescriptionStatus.WITH_DISPENSER,
            PrescriptionStatus.WITH_DISPENSER_ACTIVE,
            PrescriptionStatus.DISPENSED,
            PrescriptionStatus.NOT_DISPENSED,
            PrescriptionStatus.CLAIMED,
            PrescriptionStatus.REPEAT_DISPENSE_FUTURE_INSTANCE,
        ]
        status_filters = [Conditions.status_equals(status) for status in desired_statuses]
        filter_expressions = filter_expressions & (
            functools.reduce(lambda a, b: a | b, status_filters)
        )

        query = DynamoDbQuery(
            self.client,
            self.log_object,
            internal_id,
            GSI.NHS_NUMBER_DATE_2,
            key_conditions,
            filter_expressions,
            limit,
            descending=True,
        )

        prescription_ids = list([item["pk"] for item in query])
        more_results = not query.complete
        return more_results, prescription_ids



import datetime
import re
import uuid

from eps_spine_shared.errors import EpsSystemError
from eps_spine_shared.nhsfundamentals.timeutilities import TimeFormats


class ChangeLogProcessor(object):
    """
    Keep the change log within the record

    The methods here assume that a None is never passed as change log, if necessary pass {} instead.
    """

    TIMESTAMP = "Timestamp"
    SCN = "SCN"
    SYS_SDS = "agentSystemSDS1"
    PRS_SDS = "agentPersonSDSPerson"
    UPDATES = "updatesApplied"
    XSLT = "Source XSLT"
    RSP_PARAMS = "Response Parameters"
    NOTIFICATIONS = "Notifications"
    INTERNAL_ID = "InternalID"
    INTERACTION_ID = "interactionID"
    TIME_PREPARED = "timePreparedForUpdate"
    INSTANCE = "instance"

    RECORD_SCN_REF = "SCN"
    RECORD_CHANGELOG_REF = "changeLog"

    INITIAL_SCN = 1

    DO_NOT_PRUNE = -1
    PRUNE_POINT = 12
    INVALID_SCN = -1

    @classmethod
    def log_for_general_update(cls, scn, internal_id=None, xslt=None, rsp_parameters=None):
        """
        Add a general change log update, nothing specific to a domain
        """
        if not rsp_parameters:
            rsp_parameters = {}

        logOfChange = {}
        _timeOfChange = datetime.datetime.now().strftime(TimeFormats.STANDARD_DATE_TIME_FORMAT)
        logOfChange[cls.TIMESTAMP] = _timeOfChange
        logOfChange[cls.SCN] = scn
        logOfChange[cls.INTERNAL_ID] = internal_id
        logOfChange[cls.XSLT] = xslt
        logOfChange[cls.RSP_PARAMS] = rsp_parameters
        return logOfChange

    @classmethod
    def update_change_log(cls, record, new_log, message_id, prune_point=None):
        """
        Take a change log from the record, add the new log to it, and prune to the prune
        point
        """
        if not prune_point:
            prune_point = cls.PRUNE_POINT

        change_log = record.get(cls.RECORD_CHANGELOG_REF, {})
        change_log[message_id] = new_log
        cls.prune_change_log(change_log, prune_point)

        record[cls.RECORD_CHANGELOG_REF] = change_log
        return record

    @classmethod
    def prune_change_log(cls, change_log, prune_point):
        """
        Prune to the prune point
        """
        if prune_point != cls.DO_NOT_PRUNE:
            _, highest_scn = cls.get_highest_scn(change_log)
            if highest_scn != cls.INVALID_SCN:
                scn_to_prune = highest_scn - prune_point
                prune_list = []
                for guid, change_log_entry in change_log.items():
                    entry_scn = int(change_log_entry.get(cls.SCN, cls.INVALID_SCN))
                    if entry_scn < scn_to_prune:
                        prune_list.append(guid)
                for guid in prune_list:
                    del change_log[guid]

    @classmethod
    def get_highest_scn(cls, change_log):
        """
        Return the (guid, scn) from the first change_log found with the highest SCN
        """
        (highest_guid, highest_scn) = (None, cls.INVALID_SCN)
        for guid in change_log:
            scn = int(change_log[guid].get(cls.SCN, cls.INVALID_SCN))
            if scn > highest_scn:
                highest_guid = guid
                highest_scn = scn
        return (highest_guid, highest_scn)

    @classmethod
    def get_scn(cls, change_log_entry):
        """
        Retrieve the SCN as an int from the provided change_log entry
        """
        scn_number = int(change_log_entry.get(cls.SCN, cls.INVALID_SCN))
        return scn_number

    @classmethod
    def list_scns(cls, change_log):
        """
        Performs list comprehension on the change_log dictionary to retrieve all the SCNs from change_log

        Duplicates will be present and change_log entries with no SCN will be represented with the
        INVALID_SCN constant
        """
        scn_number_list = [cls.get_scn(change_log[x]) for x in change_log]
        return scn_number_list

    @classmethod
    def get_max_scn(cls, change_log):
        """
        Return the highest SCN value from the provided change_log
        """
        scn_number_list = cls.list_scns(change_log)
        if not scn_number_list:
            return cls.INVALID_SCN
        highest_scn = max(scn_number_list)
        return highest_scn

    @classmethod
    def get_all_guids_for_scn(cls, change_log, search_scn):
        """
        For the provided SCN return the GUID Keys of all the change_log entries that have that SCN

        Usually this will be a single GUID, but in the case of tickled records there can be multiple.
        """
        search_scn = int(search_scn)
        guid_list = [k for k in change_log if cls.get_scn(change_log[k]) == search_scn]
        return guid_list

    @classmethod
    def get_max_scn_guids(cls, change_log):
        """
        Finds the highest SCN in the change_log and returns all the GUIDs that have that SCN
        """
        highest_scn = cls.get_max_scn(change_log)
        guid_list = cls.get_all_guids_for_scn(change_log, highest_scn)
        return guid_list

    @classmethod
    def get_all_guids(cls, change_log):
        """
        Return a list of all the GUID keys from the provided change_log
        """
        return list(change_log.keys())

    @classmethod
    def get_last_change_time(cls, change_log):
        """
        Returns the last change time
        """
        try:
            guid = cls.get_max_scn_guids(change_log)[0]
        except IndexError:
            return None
        return change_log[guid].get(cls.TIMESTAMP)

    @classmethod
    def set_initial_change_log(cls, record, internal_id, reason_guid=None):
        """
        If no change log is present set an initial change log on the record.  It may
        use a GUID as a key or a string explaining the reason for initiating the
        change log.
        """
        change_log = record.get(cls.RECORD_CHANGELOG_REF)
        if change_log:
            return

        scn = int(record.get(cls.RECORD_SCN_REF, cls.INITIAL_SCN))
        if not reason_guid:
            reason_guid = str(uuid.uuid4()).upper()
        change_log = {}
        change_log[reason_guid] = cls.log_for_general_update(scn, internal_id)

        record[cls.RECORD_CHANGELOG_REF] = change_log


class DemographicsChangeLogProcessor(ChangeLogProcessor):
    """
    Change Log Processor specifically for demographic records
    """

    # Demographic record uses 'serialChangeNumber' rather than the default 'SCN'
    RECORD_SCN_REF = "serialChangeNumber"

    @classmethod
    def log_for_domain_update(cls, update_context, internal_id):
        """
        Create a change log for this expected change - requires attributes to be set on
        context object
        """
        log_of_change = cls.log_for_general_update(
            update_context.pdsRecord.get(cls.RECORD_SCN_REF, cls.INITIAL_SCN),
            internal_id,
            update_context.responseDetails.get(cls.XSLT),
            update_context.responseDetails.get(cls.RSP_PARAMS),
        )

        log_of_change[cls.SYS_SDS] = update_context.agentSystem
        log_of_change[cls.PRS_SDS] = update_context.agentPerson
        log_of_change[cls.UPDATES] = update_context.updatesApplied
        log_of_change[cls.NOTIFICATIONS] = update_context.notificationsToQueue
        return log_of_change

    @staticmethod
    def get_highest_gp_links_transaction_number(change_log, sender, recipient):
        """
        Return the highest GP Links transaction number which has been included in the change log, or None (if there
        aren't any).
        """
        max_number = -1

        gp_links_key_pattern = re.compile(
            "^{}_{}_[0-9]+_[0-9]+_(?P<transactionNumber>[0-9]+)$".format(
                sender.upper(), recipient.upper()
            )
        )

        for key in change_log.keys():  # noqa: SIM118
            match = gp_links_key_pattern.match(key)
            # Ignore keys which aren't related to GP Links transactions
            if match is None:
                continue
            transaction_number = int(match.group("transactionNumber"))
            if transaction_number > max_number:
                max_number = transaction_number

        return max_number


class PrescriptionsChangeLogProcessor(ChangeLogProcessor):
    """
    Change Log Processor specifically for prescriptions records
    """

    FROM_STATUS = "fromStatus"
    TO_STATUS = "toStatus"
    INS_FROM_STATUS = "instanceFromStatus"
    INS_TO_STATUS = "instanceToStatus"
    PRE_CHANGE_STATUS_DICT = "preChangeStatusDict"
    POST_CHANGE_STATUS_DICT = "postChangeStatusDict"
    CHANGED_ISSUES_LIST = "issuesAlteredByChange"
    PRE_CHANGE_CURRENT_ISSUE = "preChangeCurrentIssue"
    POST_CHANGE_CURRENT_ISSUE = "postChangeCurrentIssue"
    TOUCHED = "touched"
    AGENT_ROLE_PROFILE_CODE_ID = "agentRoleProfileCodeId"
    AGENT_PERSON_ROLE = "agentPersonRole"
    AGENT_PERSON_ORG_CODE = "agentPersonOrgCode"

    MIN_INITIALHISTORY = 16
    MIN_RECENTHISTORY = 16
    REPEATING_ACTIONS = [
        "PORX_IN060102UK30",
        "PORX_IN060102SM30",
        "PORX_IN132004UK30",
        "PORX_IN132004SM30",
        "PORX_IN132004UK04",
        "PORX_IN100101UK31",
        "PORX_IN100101SM31",
        "PORX_IN100101UK04",
        "PORX_IN020101UK31",
        "PORX_IN020102UK31",
        "PORX_IN020101SM31",
        "PORX_IN020102SM31",
        "PORX_IN020101UK04",
        "PORX_IN020102UK04",
        "PORX_IN060102GB01",
        "PRESCRIPTION_DISPENSE_PROPOSAL_RETURN",
    ]

    REGEX_ALPHANUMERIC8 = re.compile(r"^[A-Za-z0-9\-]{1,8}$")

    @classmethod
    def log_for_domain_update(cls, update_context, internal_id):
        """
        Create a change log for this expected change - requires attribute to be set on
        context object
        """

        log_of_change = cls.log_for_general_update(
            update_context.epsRecord.get_scn(),
            internal_id,
            update_context.responseDetails.get(cls.XSLT),
            update_context.responseDetails.get(cls.RSP_PARAMS),
        )
        log_of_change = update_context.workDescriptionObject.createInitialEventLog(log_of_change)

        instance = (
            str(update_context.updateInstance)
            if update_context.updateInstance
            else str(update_context.instanceID)
        )

        log_of_change[cls.TIME_PREPARED] = update_context.handleTime.strftime(
            TimeFormats.STANDARD_DATE_TIME_FORMAT
        )

        # NOTE: FROM_STATUS and TO_STATUS seem to be legacy fields, that have been
        # superceded by the INS_FROM_STATUS and INS_TO_STATUS fields set below.
        # The only reference to TO_STATUS seems to be in PrescriptionJsonQueryResponse.cfg
        # template used by the prescription detail view web service
        log_of_change[cls.FROM_STATUS] = (
            update_context.epsRecord.return_previous_prescription_status(
                update_context.instanceID, False
            )
        )
        log_of_change[cls.TO_STATUS] = update_context.epsRecord.return_prescription_status(
            update_context.instanceID, False
        )

        # Event history lines for UI
        # **** NOTE THAT THESE ARE WRONG, THEY REFER TO THE FINAL ISSUE, WHICH MAY NOT BE THE ISSUE THAT WAS UPDATED
        log_of_change[cls.INSTANCE] = instance
        log_of_change[cls.INS_FROM_STATUS] = (
            update_context.epsRecord.return_previous_prescription_status(instance, False)
        )
        log_of_change[cls.INS_TO_STATUS] = update_context.epsRecord.return_prescription_status(
            instance, False
        )
        log_of_change[cls.AGENT_ROLE_PROFILE_CODE_ID] = update_context.agentRoleProfileCodeId
        log_of_change[cls.AGENT_PERSON_ROLE] = update_context.agentPersonRole
        org_code = update_context.agentOrganization
        has_dispenser_code = (
            hasattr(update_context, "dispenserCode") and update_context.dispenserCode
        )
        if (
            not org_code
            and has_dispenser_code
            and cls.REGEX_ALPHANUMERIC8.match(update_context.dispenserCode)
        ):
            org_code = update_context.dispenserCode
        log_of_change[cls.AGENT_PERSON_ORG_CODE] = org_code

        # To help with troubleshooting, the following change entries are added
        pre_change_issue_statuses = update_context.epsRecord.return_prechange_issue_status_dict()
        post_change_issue_statuses = update_context.epsRecord.create_issue_current_status_dict()
        log_of_change[cls.PRE_CHANGE_STATUS_DICT] = pre_change_issue_statuses
        log_of_change[cls.POST_CHANGE_STATUS_DICT] = post_change_issue_statuses
        log_of_change[cls.CHANGED_ISSUES_LIST] = update_context.epsRecord.return_changed_issue_list(
            pre_change_issue_statuses,
            post_change_issue_statuses,
            None,
            update_context.changedIssuesList,
        )
        # To help with troubleshooting, the following currentIssue values are added
        log_of_change[cls.PRE_CHANGE_CURRENT_ISSUE] = (
            update_context.epsRecord.return_prechange_current_issue()
        )
        log_of_change[cls.POST_CHANGE_CURRENT_ISSUE] = update_context.epsRecord.current_issue_number
        if hasattr(update_context, cls.TOUCHED) and update_context.touched:
            log_of_change[cls.TOUCHED] = update_context.touched

        return log_of_change

    @classmethod
    def prune_change_log(cls, change_log, prune_point):
        """
        Prune the change log where there is a series of change log entries for the same
        interactionID - and the change is neither recent nor part of the early history

        The intention is that if we get a repeating interaction we don't continue to explode the
        changeLog with all the history
        """
        inverted_change_log = {}
        max_scn = 0
        for guid, change_log_entry in change_log.items():
            scn = int(change_log_entry.get(cls.SCN, cls.INVALID_SCN))
            inverted_change_log[scn] = (guid, change_log_entry.get(cls.INTERACTION_ID))
            max_scn = max(max_scn, scn)
        if max_scn <= prune_point:
            # Don't make any changes
            return

        icl_scn_keys = list(inverted_change_log.keys())
        icl_scn_keys.sort(reverse=True)
        guids_to_prune = []
        for icl_scn in icl_scn_keys:
            if icl_scn > (max_scn - cls.MIN_RECENTHISTORY) or icl_scn < cls.MIN_INITIALHISTORY:
                continue
            this_int_id = inverted_change_log.get(icl_scn, (None, None))[1]
            (previous_guid, previous_int_id) = inverted_change_log.get(icl_scn - 1, (None, None))
            one_before_int_id = inverted_change_log.get(icl_scn - 2, (None, None))[1]
            if (
                this_int_id
                and this_int_id in cls.REPEATING_ACTIONS
                and this_int_id == previous_int_id
                and previous_int_id == one_before_int_id
            ):
                guids_to_prune.append(previous_guid)

        for guid in guids_to_prune:
            del change_log[guid]

        if len(change_log) > prune_point:
            # If we have breached the prune point but can't safely prune - stop before
            # The un-pruned record becomes an issue
            raise EpsSystemError(EpsSystemError.SYSTEM_FAILURE)



import random
from decimal import Decimal
from enum import Enum
from typing import Optional


class ReleaseVersion(Enum):
    """
    Enum of release versions to be used in the DynamoDB table.
    """

    R1 = "R1"
    R2 = "R2"
    UNKNOWN = "UNKNOWN"


class RecordType(Enum):
    """
    Enum of release types to be used in the DynamoDB table.
    """

    ACUTE = "Acute"
    REPEAT_PRESCRIBE = "RepeatPrescribe"
    REPEAT_DISPENSE = "RepeatDispense"


class DefinedAttributeType(Enum):
    """
    S/N type for a defined attribute
    """

    STRING = "S"
    NUMBER = "N"


class DefinedAttribute:
    """
    Definition of an attribute in the DynamoDB table.
    """

    def __init__(self, name: str, arg_type: DefinedAttributeType) -> None:
        self.name = name
        self.type = arg_type


class Key(Enum):
    """
    Enum of table Keys
    """

    PK = DefinedAttribute("pk", DefinedAttributeType.STRING)
    SK = DefinedAttribute("sk", DefinedAttributeType.STRING)

    @property
    def name(self) -> str:
        return self.value.name

    @property
    def attribute_type(self) -> DefinedAttributeType:
        return self.value.type


class Attribute(Enum):
    """
    Enum of Defined Attributes to be used in the DynamoDB table.
    """

    NHS_NUMBER = DefinedAttribute("nhsNumber", DefinedAttributeType.STRING)
    CREATION_DATETIME = DefinedAttribute("creationDatetime", DefinedAttributeType.STRING)
    PRESCRIBER_ORG = DefinedAttribute("prescriberOrg", DefinedAttributeType.STRING)
    DISPENSER_ORG = DefinedAttribute("dispenserOrg", DefinedAttributeType.STRING)
    NOMINATED_PHARMACY = DefinedAttribute("nominatedPharmacy", DefinedAttributeType.STRING)
    IS_READY = DefinedAttribute("isReady", DefinedAttributeType.NUMBER)
    NEXT_ACTIVITY = DefinedAttribute("nextActivity", DefinedAttributeType.STRING)
    NEXT_ACTIVITY_DATE = DefinedAttribute("nextActivityDate", DefinedAttributeType.STRING)
    DOC_REF_TITLE = DefinedAttribute("docRefTitle", DefinedAttributeType.STRING)
    CLAIM_NOTIFICATION_STORE_DATE = DefinedAttribute(
        "claimNotificationStoreDate", DefinedAttributeType.STRING
    )
    STORE_TIME = DefinedAttribute("storeTime", DefinedAttributeType.STRING)
    BACKSTOP_DELETE_DATE = DefinedAttribute("backstopDeleteDate", DefinedAttributeType.STRING)
    SEQUENCE_NUMBER = DefinedAttribute("sequenceNumber", DefinedAttributeType.NUMBER)
    SEQUENCE_NUMBER_NWSSP = DefinedAttribute("sequenceNumberNwssp", DefinedAttributeType.NUMBER)
    LM_DAY = DefinedAttribute("_lm_day", DefinedAttributeType.STRING)
    RIAK_LM = DefinedAttribute("_riak_lm", DefinedAttributeType.NUMBER)
    BATCH_CLAIM_ID = DefinedAttribute("batchClaimId", DefinedAttributeType.STRING)

    @property
    def name(self) -> str:
        return self.value.name

    @property
    def attribute_type(self) -> DefinedAttributeType:
        return self.value.type


class ProjectedAttribute(Enum):
    """
    Enum of Projected Attributes to be used in the DynamoDB table.
    """

    CLAIM_IDS = "claimIds"
    SCN = "scn"
    STATUS = "status"
    BODY = "body"
    INDEXES = "indexes"
    EXPIRE_AT = "expireAt"
    RELEASE_VERSION = "releaseVersion"
    RECORD_TYPE = "recordType"

    @property
    def name(self) -> str:
        return self.value


class SortKey(Enum):
    """
    Enum of SortKeys to be used in the DynamoDB table.
    """

    DOCUMENT = "DOC"
    RECORD = "REC"
    WORK_LIST = "WRK"
    CLAIM = "CLM"
    SEQUENCE_NUMBER = "SQN"


class Index:
    """
    Information on a GSI
    """

    def __init__(self, name: str, pk: Attribute, sk: Optional[Attribute]) -> None:
        self.name = name
        self.pk = pk
        self.sk = sk if sk else None


class GSI(Enum):
    """
    Enum of global secondary indexes of the DynamoDB table.
    """

    NHS_NUMBER_DATE = Index("nhsNumberDate", Attribute.NHS_NUMBER, Attribute.CREATION_DATETIME)
    NHS_NUMBER_DATE_2 = Index("nhsNumberDate2", Attribute.NHS_NUMBER, Attribute.CREATION_DATETIME)
    PRESCRIBER_DATE = Index("prescriberDate", Attribute.PRESCRIBER_ORG, Attribute.CREATION_DATETIME)
    DISPENSER_DATE = Index("dispenserDate", Attribute.DISPENSER_ORG, Attribute.CREATION_DATETIME)
    NOMINATED_PHARMACY_STATUS = Index(
        "nominatedPharmacyStatus", Attribute.NOMINATED_PHARMACY, Attribute.IS_READY
    )
    CLAIM_ID = Index("claimId", Key.SK, Attribute.BATCH_CLAIM_ID)
    NEXT_ACTIVITY_DATE = Index(
        "nextActivityDate", Attribute.NEXT_ACTIVITY, Attribute.NEXT_ACTIVITY_DATE
    )
    STORE_TIME_DOC_REF_TITLE = Index(
        "storeTimeDocRefTitle", Attribute.DOC_REF_TITLE, Attribute.STORE_TIME
    )
    CLAIM_NOTIFICATION_STORE_TIME = Index(
        "claimNotificationStoreTime", Attribute.CLAIM_NOTIFICATION_STORE_DATE, Attribute.STORE_TIME
    )
    BACKSTOP_DELETE_DATE = Index("backstopDeleteDate", Key.SK, Attribute.BACKSTOP_DELETE_DATE)
    CLAIM_ID_SEQUENCE_NUMBER = Index("claimIdSequenceNumber", Attribute.SEQUENCE_NUMBER, None)
    CLAIM_ID_SEQUENCE_NUMBER_NWSSP = Index(
        "claimIdSequenceNumberNwssp", Attribute.SEQUENCE_NUMBER_NWSSP, None
    )
    LAST_MODIFIED = Index("lastModified", Attribute.LM_DAY, Attribute.RIAK_LM)

    @property
    def name(self) -> str:
        return self.value.name

    @property
    def pk(self) -> Attribute:
        return self.value.pk

    @property
    def sk(self) -> Optional[Attribute]:
        return self.value.sk


REGION_NAME = "eu-west-2"
SERVICE_NAME = "dynamodb"
CONDITION_EXPRESSION = (
    f"attribute_not_exists({Key.PK.name}) AND attribute_not_exists({Key.SK.name})"
)
LAST_MODIFIED_DAILY_PARTITIONS = 12
NEXT_ACTIVITY_DATE_PARTITIONS = 12
RELEASE_VERSION_PARTITIONS = 12


def replace_decimals(obj):
    """
    Utility function to replace any instances of Decimal type with int/float.
    """

    def handle_decimal(obj):
        return int(obj) if obj % 1 == 0 else float(obj)

    def handle_dict(obj):
        for k in obj:
            obj[k] = replace_decimals(obj[k])
        return obj

    def handle_list(obj):
        for i in range(len(obj)):
            obj[i] = replace_decimals(obj[i])
        return obj

    handlers = {Decimal: handle_decimal, dict: handle_dict, list: handle_list}

    return handlers.get(type(obj), lambda obj: obj)(obj)


def prescription_id_without_check_digit(prescription_id) -> str:
    """
    If length is > 36 then long prescription id with checksum so truncate to 36 characters.
    If length is > 19 and < 36 then short prescription id with checksum so truncate to 19 characters.
    """
    if len(prescription_id) > 36:
        return prescription_id[:36]
    elif len(prescription_id) > 19 and len(prescription_id) < 36:
        return prescription_id[:19]
    else:
        return prescription_id


def determine_release_version(prescription_id) -> str:
    """
    Determines the release version of a prescription based on the length of its id. Includes shard for indexing.
    """
    id_length = len(prescription_id_without_check_digit(prescription_id))
    shard = random.randint(1, RELEASE_VERSION_PARTITIONS)
    match id_length:
        case 36:
            return f"{ReleaseVersion.R1.value}.{shard}"
        case 19:
            return f"{ReleaseVersion.R2.value}.{shard}"
        case _:
            return ReleaseVersion.UNKNOWN.value
